# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now export'command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from argparse import Namespace

from ..persistence.models.graphs.dependency_graph import DependencyConfig
from ..persistence.models import Trial
from ..persistence import persistence_config


from .command import NotebookCommand
from .cmd_diff import Diff
from .cmd_history import History
from .cmd_show import Show


class Export(NotebookCommand):
    """Export the collected provenance of a trial to Prolog or Notebook"""

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("-r", "--rules", action="store_true",
                help="also exports inference rules")
        DependencyConfig.create_arguments(add_arg, mode="simulation")
        add_arg("trial", type=str, nargs="?",
                help="trial id or none for last trial. If you are generation "
                     "ipython notebook files, it is also possible to use "
                     "'history' or 'diff:<trial_id_1>:<trial_id_2>'")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        trial = Trial(trial_ref=args.trial)
        trial.dependency_config.read_args(args)
        print(trial.prolog.export_text_facts())
        if args.rules:
            print("\n".join(trial.prolog.rules()))

    def execute_export(self, args):
        namespace = Namespace(ipynb=True, dir=args.dir)
        if not args.trial or args.trial == "current":
            namespace.trial = None
            return Show().execute_export(namespace)
        if args.trial == "history":
            return History().execute_export(namespace)

        splitted = dict(enumerate(args.trial.split(":")))
        if splitted[0] == "diff":
            namespace.trial1 = splitted.get(1)
            namespace.trial2 = splitted.get(2)
            return Diff().execute_export(namespace)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now run' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse
import os
import sys

from ..collection.metadata import Metascript
from ..persistence.models import Tag, Trial
from ..utils import io, metaprofiler
from ..utils.cross_version import PY3

from .command import Command


def non_negative(string):
    """Check if argument is >= 0"""
    value = int(string)
    if value < 0:
        raise argparse.ArgumentTypeError(
            "{} is not a non-negative integer value".format(string))
    return value


class ScriptArgs(argparse.Action):                                               # pylint: disable=too-few-public-methods
    """Action to create script attribute"""
    def __call__(self, parser, namespace, values, option_string=None):
        if not values:
            raise argparse.ArgumentError(
                self, "can't be empty")

        script = os.path.realpath(values[0])

        if not os.path.exists(script):
            raise argparse.ArgumentError(
                self, "can't open file '{}': "
                "[Errno 2] No such file or directory".format(values[0]))

        setattr(namespace, self.dest, script)
        setattr(namespace, "argv", values)


def run(metascript):
    """Execute noWokflow to capture provenance from script"""
    try:
        metascript.trial_id = Trial.store(*metascript.create_trial_args())
        Tag.create_automatic_tag(*metascript.create_automatic_tag_args())

        io.print_msg("collecting definition provenance")
        metascript.definition.collect_provenance()
        metascript.definition.store_provenance()

        io.print_msg("collecting deployment provenance")
        metascript.deployment.collect_provenance()
        metascript.deployment.store_provenance()

        io.print_msg("collection execution provenance")
        metascript.execution.collect_provenance()
        metascript.execution.store_provenance()

        metaprofiler.meta_profiler.save()

    finally:
        metascript.create_last()


class Run(Command):
    """Run a script collecting its provenance"""

    def __init__(self, *args, **kwargs):
        super(Run, self).__init__(*args, **kwargs)
        self.default_context = "main"
        self.default_call_storage_frequency = 10000
        self.default_save_frequency = 0
        self.default_execution_provenance = "Profiler"
        self.add_help = False

    def add_arguments(self):
        add_arg = self.add_argument
        add_cmd = self.add_argument_cmd

        # It will create both script and argv var
        add_cmd("script", nargs=argparse.REMAINDER, action=ScriptArgs,
                help="Python script to be executed")

        # Deployment
        if not self.is_ipython:
            deployment = self.parser.add_argument_group(
                "optional deployment arguments")
            add_arg = deployment.add_argument
        add_arg("-b", "--bypass-modules", action="store_true",
                help="bypass module dependencies analysis, assuming that no "
                     "module changes occurred since last execution")

        # Execution
        if not self.is_ipython:
            execution = self.parser.add_argument_group(
                "optional execution arguments")
            add_arg = execution.add_argument
        add_arg("-d", "--depth", type=non_negative,
                default=sys.getrecursionlimit(),
                help="depth for capturing function activations (default: "
                     "recursion limit)")
        add_arg("-D", "--non-user-depth", type=non_negative, default=1,
                help="depth for capturing function activations outside the "
                     "selected context (default: 1)")
        add_arg("-e", "--execution-provenance",
                default=self.default_execution_provenance,
                choices=["Profiler", "Tracer", "Tracker"],
                help="R|execution provenance provider. (default: Profiler)\n"
                     "Profiler captures function calls, parameters, file \n"
                     "accesses, and globals. \n"
                     "Tracker captures everything the Profiler captures, \n"
                     "in addition to variables and dependencies.\n"
                     "Tracer is an alias to Tracker")
        add_arg("-c", "--context", choices=["main", "package", "all"],
                default=self.default_context,
                help="functions subject to depth computation when capturing "
                     "activations (default: main)")
        add_arg("-s", "--save-frequency", type=non_negative,
                default=self.default_save_frequency,
                help="frequency (in ms) to save partial provenance")
        add_arg("-S", "--call-storage-frequency", type=non_negative,
                default=self.default_call_storage_frequency,
                help="frequency (in calls) to save partial provenance")

        # Other
        if not self.is_ipython:
            other = self.parser.add_argument_group(
                "other optional arguments")
            add_arg = other.add_argument
        add_arg("-h", "--help", action="help",
                help="show this help message and exit")
        add_arg("--name", type=str,
                help="R|set script name.\n"
                     "This options allows separating or grouping trials \n"
                     "in disjoint branches in the history. It is specially\n"
                     "useful from grouping Jupyter Cells")
        add_arg("--dir", type=str,
                help="set project path. The noworkflow database folder will "
                     "be created in this path. Default to script directory")
        add_arg("-v", "--verbose", action="store_true",
                help="increase output verbosity")

        # Internal
        add_arg("--disasm0", action="store_true", help=argparse.SUPPRESS)
        add_arg("--disasm", action="store_true", help=argparse.SUPPRESS)
        add_cmd("--create_last", action="store_true", help=argparse.SUPPRESS)
        add_arg("--meta", action="store_true", help=argparse.SUPPRESS)

    def execute(self, args):
        if args.meta:
            metaprofiler.meta_profiler.active = True
            metaprofiler.meta_profiler.data["cmd"] = " ".join(sys.argv)

        io.verbose = args.verbose
        io.print_msg("removing noWorkflow boilerplate")

        # Create Metascript with params
        metascript = Metascript().read_cmd_args(args)

        # Set __main__ namespace
        import __main__
        metascript.namespace = __main__.__dict__

        # Clear boilerplate
        metascript.clear_sys()
        metascript.clear_namespace()

        # Run script
        run(metascript)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""IPython Module"""
from __future__ import (absolute_import, print_function,
                        division)


from ..persistence.models import *                                               # pylint: disable=wildcard-import
from ..persistence import persistence_config, relational, content


def init(path=None, ipython=None):
    """Initiate noWorkflow extension.
    Load D3, IPython magics, and connect to database


    Keyword Arguments:
    path -- database path (default=current directory)
    ipython -- IPython object (default=None)
    """

    import os
    from IPython.display import display_html
    from ..utils.functions import resource
    from .magics import register_magics
    from .hierarchymagic import load_ipython_extension as load_hierarchy

    register_magics(ipython)
    load_hierarchy(ipython)

    if path is None:
        path = os.getcwd()
    persistence_config.connect(path)

    js_files = [
        u"vis/static/d3-v3.4.11/d3.min.js",
        u"vis/static/trial_graph.js",
        u"vis/static/history_graph.js",
        u"vis/static/ipython.js",
    ]

    require_js = u"""
        <!DOCTYPE html>
        <meta charset="utf-8">
        <body>
            <script charset="utf-8">
            var now_temp = define;
            define = undefined;
            {0}
            define = now_temp;
            </script>
        </body
    """
    js_text = [resource(js_file, u"utf-8") for js_file in js_files]

    display_html(
        require_js.format(";\n".join(js_text)),
        raw=True
    )

    css_files = [
        u"vis/static/font-awesome-4.3.0/css/font-awesome-ip.css",
        u"vis/static/shared_graph.css",
        u"vis/static/trial_graph.css",
        u"vis/static/history_graph.css",
    ]

    css_lines = [u"<style>"]
    for css_file in css_files:
        css_lines.append(resource(css_file, u"utf-8"))
    css_lines.append(u"</style>")
    display_html(u"\n".join(css_lines), raw=True)

    return u"ok"



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

# Do not add from __future__ imports here
"""Provide support for both Python 2 and Python 3"""

import sys
import numbers

try:
    from cStringIO import StringIO                                               # pylint: disable=unused-import
except ImportError:
    from io import StringIO


PY3 = (sys.version_info >= (3, 0))
if PY3:
    import builtins                                                              # pylint: disable=wrong-import-position, unused-import
    import pickle                                                                # pylint: disable=wrong-import-position, unused-import
    import reprlib                                                               # pylint: disable=wrong-import-position, unused-import
    from itertools import zip_longest                                            # pylint: disable=wrong-import-position, unused-import

    IMMUTABLE = (None.__class__, bool, numbers.Number, str, bytes)
    string = (str, bytes)                                                        # pylint: disable=invalid-name
    raw_bytes = (bytes, bytearray)                                               # pylint: disable=invalid-name
else:
    import __builtin__ as builtins                                               # pylint: disable=wrong-import-position, unused-import, import-error
    try:
        import cPickle as pickle                                                 # pylint: disable=wrong-import-position, unused-import
    except ImportError:
        import pickle                                                            # pylint: disable=wrong-import-position, unused-import, ungrouped-imports
    import repr as reprlib                                                       # pylint: disable=wrong-import-position, unused-import, import-error
    from itertools import izip_longest as zip_longest                            # pylint: disable=wrong-import-position, unused-import, ungrouped-imports

    IMMUTABLE = (None.__class__, bool, numbers.Number, basestring)               # pylint: disable=invalid-name, undefined-variable
    string = (basestring,)                                                       # pylint: disable=invalid-name, undefined-variable
    raw_bytes = (str,)                                                           # pylint: disable=invalid-name


def cross_compile(*args, **kwargs):
    """Compile the source string into a code object

    __future__ imports change the behavior of default compile function
    This function just provides the 'compile' free of __future__ imports
    """
    return compile(*args, **kwargs)


def bytes_string(text, encode="utf-8"):
    """Return a bytes object on Python 3 and a str object on Python 2"""
    if not PY3:
        if isinstance(text, unicode):                                            # pylint: disable=undefined-variable
            result = text.encode(encode)
        else:
            result = text
    else:
        if isinstance(text, bytes):
            result = text
        else:
            result = bytes(text, encode)
    return result


def default_string(text, encode="utf-8"):
    """Return a unicode object on Python 3 and a bytes object on Python 2"""
    if not PY3:
        if isinstance(text, unicode):                                            # pylint: disable=undefined-variable
            result = text.encode(encode)
        else:
            result = text
    else:
        if isinstance(text, bytes):
            result = text.decode(encode)
        else:
            result = text
    return result



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now vis' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import threading
import webbrowser

from ..persistence import persistence_config

from .command import Command


def run(path=None, browser=False, port=5000, debug=False):
    """Open Flask server"""
    if browser:
        url = "http://127.0.0.1:{0}".format(port)
        print(url)
        threading.Timer(1.25, lambda: webbrowser.open(url)).start()
    from ..vis.views import app
    app.dir = path or os.getcwd()
    app.run(port=port, debug=debug, threaded=True)


class Vis(Command):
    """ Open visualization tool """

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("-p", "--port", nargs="?", type=int, default=5000,
                help="sets server port")
        add_arg("-d", "--debug", action="store_true",
                help="debug mode")
        add_arg("-b", "--browser", action="store_true",
                help="opens browser")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        run(path=args.dir, browser=args.browser, port=args.port,
            debug=args.debug)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Slicing Variable Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text, TIMESTAMP, select, alias
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint
from sqlalchemy.orm import aliased

from .. import relational

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr, PrologTimestamp, PrologNullableRepr

from .base import AlchemyProxy, proxy_class, many_ref, many_viewonly_ref, proxy
from .base import backref_one, backref_many
from .variable_dependency import VariableDependency


@proxy_class
class Variable(AlchemyProxy):
    """Represent a variable captured during program slicing"""

    __tablename__ = "variable"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "activation_id", "id"),
        ForeignKeyConstraint(["trial_id", "activation_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    activation_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    line = Column(Integer)
    value = Column(Text)
    time = Column(TIMESTAMP)
    type = Column(Text)                                                          # pylint: disable=invalid-name

    usages = many_ref("variable", "VariableUsage")

    # dependencies in which this variable is the dependent
    dependencies_as_source = many_viewonly_ref(
        "source", "VariableDependency",
        primaryjoin=(
            (id == VariableDependency.m.source_id) &
            (activation_id == VariableDependency.m.source_activation_id) &
            (trial_id == VariableDependency.m.trial_id))
    )

    # dependencies in which this variable is the dependency
    dependencies_as_target = many_viewonly_ref(
        "target", "VariableDependency",
        primaryjoin=(
            (id == VariableDependency.m.target_id) &
            (activation_id == VariableDependency.m.target_activation_id) &
            (trial_id == VariableDependency.m.trial_id)))

    dependencies = many_viewonly_ref(
        "dependents", "Variable",
        secondary=VariableDependency.__table__,
        primaryjoin=(
            (id == VariableDependency.m.source_id) &
            (activation_id == VariableDependency.m.source_activation_id) &
            (trial_id == VariableDependency.m.trial_id)),
        secondaryjoin=(
            (id == VariableDependency.m.target_id) &
            (activation_id == VariableDependency.m.target_activation_id) &
            (trial_id == VariableDependency.m.trial_id)))

    trial = backref_one("trial")  # Trial.variables
    activation = backref_one("activation")  # Activation.variables
    dependents = backref_many("dependents")  # Variable.dependencies

    prolog_description = PrologDescription("variable", (
        PrologTrial("trial_id", link="activation.trial_id"),
        PrologAttribute("activation_id", link="activation.id"),
        PrologAttribute("id"),
        PrologRepr("name"),
        PrologAttribute("line"),
        PrologNullableRepr("value"),
        PrologTimestamp("timestamp", attr_name="time"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "during a specific function activation (*activation_id*),\n"
        "in a specific *line* of code,\n"
        "and in a specific *timestamp*,\n"
        "a variable *name* was updated\n"
        "to a new *value*."
    ))


    def __key(self):
        return self.id

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if not isinstance(other, Variable):
            return False
        return self.__key() == other.__key()                                     # pylint: disable=protected-access

    @property
    def return_dependency(self):
        """Return "return" dependency. Valid only for call type"""
        return proxy(self._get_instance().dependencies.filter(
            Variable.m.name == "return").first())

    @property
    def box_dependency(self):
        """Return "return" dependency. Valid only for call type"""
        return proxy(self._get_instance().dependencies.filter(
            Variable.m.name.like("%box--")).first())

    @classmethod  # query
    def fast_arg_and_original(cls, trial_id, session=None):
        """Return tuples with variable of type arg and original variable"""
        # ToDo: optimize here
        session = session or relational.session
        avar = alias(cls.t, name="avar")
        abox = alias(VariableDependency.t, name="abox")
        box = alias(cls.t, name="box")
        obox = alias(VariableDependency.t, name="obox")
        ovar = alias(cls.t, name="ovar")


        query = select([avar.c.id, ovar.c.id]).select_from(
            ovar.join(obox, (
                (ovar.c.id == obox.c.target_id) &
                (ovar.c.trial_id == obox.c.trial_id) &
                (ovar.c.activation_id == obox.c.target_activation_id)
            )).join(box, (
                (box.c.id == obox.c.source_id) &
                (box.c.trial_id == obox.c.trial_id) &
                (box.c.activation_id == obox.c.source_activation_id)
            )).join(abox, (
                (box.c.id == abox.c.target_id) &
                (box.c.trial_id == abox.c.trial_id) &
                (box.c.activation_id == abox.c.target_activation_id)
            )).join(avar, (
                (avar.c.id == abox.c.source_id) &
                (avar.c.trial_id == abox.c.trial_id) &
                (avar.c.activation_id == abox.c.source_activation_id)
            ))
        ).where(
            (avar.c.type == "arg") &
            (avar.c.trial_id == trial_id) &
            (ovar.c.name == avar.c.name)
        )
        return session.execute(query)

    def __repr__(self):
        return (
            "Variable({0.trial_id}, {0.activation_id}, "
            "{0.id}, {0.name}, {0.line}, {0.type})"
        ).format(self)

    def __str__(self):
        return "(L{0.line}, {0.name}, {0.value})".format(self)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Lightweight objects for storage during collection"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from datetime import datetime

from future.utils import viewitems, viewvalues

from . import content


class ObjectStore(object):
    """Temporary storage for LW objects"""

    def __init__(self, cls):
        """Initialize Object Store


        Arguments:
        cls -- LW object class
        """
        self.cls = cls
        self.store = {}
        self.id = 0                                                              # pylint: disable=invalid-name
        self.count = 0

    def __getitem__(self, index):
        return self.store[index]

    def __delitem__(self, index):
        self.store[index] = None
        self.count -= 1

    def add(self, *args):
        """Add object using its __init__ arguments and return id"""
        self.id += 1
        self.count += 1
        self.store[self.id] = self.cls(self.id, *args)
        return self.id

    def add_object(self, *args):
        """Add object using its __init__ arguments and return object"""
        self.id += 1
        self.count += 1
        self.store[self.id] = self.cls(self.id, *args)
        return self.store[self.id]

    def dry_add(self, *args):
        """Return object that would be added by add_object
        Do not add it to storage
        """
        return self.cls(-1, *args)

    def remove(self, value):
        """Remove object from storage"""
        for key, val in viewitems(self.store):
            if val == value:
                del self.store[key]
                self.count -= 1

    def __iter__(self):
        """Iterate on objects, and not ids"""
        return viewvalues(self.store)

    def items(self):
        """Iterate on both ids and objects"""
        for key, value in viewitems(self.store):
            yield key, value

    def iteritems(self):
        """Iterate on both ids and objects"""
        for key, value in viewitems(self.store):
            yield key, value

    def values(self):
        """Iterate on objects if they exist"""
        for value in viewvalues(self.store):
            if value is not None:
                yield value

    def clear(self):
        """Remove deleted objects from storage"""
        self.store = {key: val for key, val in viewitems(self.store) if val}
        self.count = len(self.store)

    def generator(self, trial_id, partial=False):
        """Generator used for storing objects in database"""
        for obj in self.values():
            if partial and obj.is_complete():
                del self[obj.id]
            obj.trial_id = trial_id
            yield obj
        if partial:
            self.clear()

    def has_items(self):
        """Return true if it has items"""
        return bool(self.count)


def define_attrs(required, extra=[]):                                            # pylint: disable=dangerous-default-value
    """Create __slots__ by adding extra attributes to required ones"""
    slots = tuple(required + extra)
    attributes = tuple(required)

    return slots, attributes


class BaseLW:                                                                    # pylint: disable=too-few-public-methods
    """Lightweight modules base class"""

    def keys(self):
        """Return attributes that should be saved"""
        return self.attributes                                                   # pylint: disable=no-member

    def __iter__(self):
        return iter(self.attributes)                                             # pylint: disable=no-member

    def __getitem__(self, key):
        if key in self.special and getattr(self, key) == -1:                     # pylint: disable=no-member
            return None
        return getattr(self, key)


# Deployment

class ModuleLW(BaseLW):
    """Module lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["id", "name", "path", "version", "code_hash"],
        ["trial_id"]
    )
    special = set()

    def __init__(self, oid, name, version, path, code_hash):                     # pylint: disable=too-many-arguments
        self.trial_id = -1
        self.id = oid                                                            # pylint: disable=invalid-name
        self.name = name
        self.version = version
        self.path = path
        self.code_hash = code_hash

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """Module can always be removed from object store"""
        return True

    def __repr__(self):
        return ("Module(id={}, name={}, version={})").format(
            self.id, self.name, self.version)


class DependencyLW(BaseLW):
    """Dependency lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["trial_id", "module_id"], ["id"]
    )
    special = set()

    def __init__(self, oid, module_id):
        self.trial_id = -1
        self.id = oid                                                            # pylint: disable=invalid-name
        self.module_id = module_id

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """Dependency can always be removed from object store"""
        return True

    def __repr__(self):
        return ("Dependency(module_id={})").format(self.module_id)


class EnvironmentAttrLW(BaseLW):
    """EnvironmentAttr lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["trial_id", "id", "name", "value"]
    )
    special = set()

    def __init__(self, oid, name, value):
        self.trial_id = -1
        self.id = oid                                                            # pylint: disable=invalid-name
        self.name = name
        self.value = value

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """EnvironmentAttr can always be removed from object store"""
        return True

    def __repr__(self):
        return ("EnvironmentAttr(id={}, name={}, value={})").format(
            self.id, self.name, self.value)


# Definition

class DefinitionLW(BaseLW):                                                      # pylint: disable=too-many-instance-attributes
    """Definition lightweight object
    May represent files, classes and function definitions
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["id", "name", "code_hash", "trial_id", "first_line", "last_line",
         "docstring"],
        ["type", "code", "parent", "namespace"],
    )
    special = set()

    def __init__(self, aid, previous_namespace, name, code, dtype, parent,       # pylint: disable=too-many-arguments
                 first_line, last_line, docstring):
        self.trial_id = -1
        self.id = aid                                                            # pylint: disable=invalid-name
        self.namespace = (
            previous_namespace +
            ("." if previous_namespace else "") +
            name
        )
        self.name = self.namespace
        self.parent = (parent if parent is not None else -1)
        self.type = dtype
        self.code = code
        self.code_hash = content.put(code.encode("utf-8"))
        self.first_line = first_line
        self.last_line = last_line
        self.docstring = docstring or ""

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """DefinitionLW can always be removed from object store"""
        return True

    def __repr__(self):
        return ("DefinitionLW(id={}, name={}, type={})").format(
            self.id, self.name, self.type)


class ObjectLW(BaseLW):
    """Object lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["trial_id", "id", "name", "type", "function_def_id"]
    )
    special = set()

    def __init__(self, oid, name, otype, function_def_id):
        self.trial_id = -1
        self.id = oid                                                            # pylint: disable=invalid-name
        self.name = name
        self.type = otype
        self.function_def_id = function_def_id

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """Object can always be removed from object store"""
        return True

    def __repr__(self):
        return (
            "Object(id={}, name={}, type={}, "
            "function_def={})"
        ).format(self.id, self.name, self.type, self.function_def_id)


# Profiler

class ActivationLW(BaseLW):                                                      # pylint: disable=too-many-instance-attributes
    """Activation lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["id", "name", "line", "return_value", "start", "finish", "caller_id",
         "trial_id"],
        ["file_accesses", "context", "slice_stack", "lasti", "definition_file",
         "args", "kwargs", "starargs", "with_definition", "filename",
         "is_main", "has_parameters",
         "loops", "conditions", "permanent_conditions",
         "temp_context", "temp_line"]
    )
    special = {"caller_id"}

    def __init__(self, aid, definition_file, filename, name, line, lasti,        # pylint: disable=too-many-arguments
                 caller_id, with_definition):
        self.trial_id = aid
        self.id = aid                                                            # pylint: disable=invalid-name
        self.name = name
        self.line = line
        self.start = datetime.now()
        self.finish = None
        self.caller_id = (caller_id if caller_id else -1)
        self.return_value = None

        # Name of the script with the call
        self.filename = filename
        # Name of the script with the function definition
        self.definition_file = definition_file
        # Activation has definition or not
        self.with_definition = with_definition
        # Activation is __main__
        self.is_main = aid == 1
        # Activation has parameters. Use only for slicing!
        self.has_parameters = True

        # File accesses. Used to get the content after the activation
        self.file_accesses = []
        # Variable context. Used in the slicing lookup
        self.context = {}

        # Temporary variables
        self.temp_context = set()
        self.temp_line = None

        # Line execution stack.
        # Used to evaluate function calls before execution line
        self.slice_stack = []
        self.lasti = lasti

        self.args = []
        self.kwargs = []
        self.starargs = []

        self.loops = []
        self.conditions = []
        self.permanent_conditions = []

    def is_complete(self):
        """Activation can be removed from object store after setting finish"""
        return self.finish is not None

    def is_comprehension(self):
        """Check if activation is comprehension"""
        return self.name in [
            "<setcomp>", "<dictcomp>", "<genexpr>", "<listcomp>"
        ]

    def __repr__(self):
        return (
            "Activation(id={}, line={}, lasti={}, filename={}, "
            " name={}, start={}, finish={}, return={}, caller_id={})"
        ).format(
            self.id, self.line, self.lasti, self.filename, self.name,
            self.start, self.finish, self.return_value, self.caller_id
        )


class ObjectValueLW(BaseLW):
    """ObjectValue lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["trial_id", "id", "name", "value", "type", "function_activation_id"]
    )
    special = set()

    def __init__(self, oid, name, value, otype, function_activation_id):         # pylint: disable=too-many-arguments
        self.trial_id = -1
        self.id = oid                                                            # pylint: disable=invalid-name
        self.name = name
        self.value = value
        self.type = otype
        self.function_activation_id = function_activation_id

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """ObjectValue can always be removed"""
        return True

    def __repr__(self):
        return (
            "ObjectValue(id={}, name={}, value={}, type={}, "
            "activation={})"
        ).format(
            self.id, self.name,
            self.value, self.type, self.function_activation_id
        )


class FileAccessLW(BaseLW):                                                      # pylint: disable=too-many-instance-attributes
    """FileAccess lightweight object
    There are type definitions on lightweight.pxd
    """

    __slots__, attributes = define_attrs(
        ["id", "name", "mode", "buffering", "timestamp", "trial_id",
         "content_hash_before", "content_hash_after",
         "function_activation_id"],
        ["done"]
    )
    special = {"function_activation_id"}

    def __init__(self, fid, name):
        self.trial_id = -1
        self.id = fid                                                            # pylint: disable=invalid-name
        self.name = name
        self.mode = "r"
        self.buffering = "default"
        self.content_hash_before = None
        self.content_hash_after = None
        self.timestamp = datetime.now()
        self.function_activation_id = -1
        self.done = False

    def update(self, variables):
        """Update file access with dict"""
        for key, value in viewitems(variables):
            setattr(self, key, value)

    def is_complete(self):
        """FileAccess can be removed once it is tagged as done"""
        return self.done

    def __repr__(self):
        return ("FileAccess(id={}, name={}").format(self.id, self.name)


# Slicing

class VariableLW(BaseLW):
    """Variable lightweight object
    There are type definitions on lightweight.pxd
    """
    __slots__, attributes = define_attrs(
        ["id", "activation_id", "name", "line", "value", "time", "trial_id",
         "type"]
    )
    special = set()

    def __init__(self, vid, activation_id, name, line, value, time, _type):             # pylint: disable=too-many-arguments
        self.id = vid                                                            # pylint: disable=invalid-name
        self.activation_id = activation_id
        self.name = name
        self.line = line
        self.value = value
        self.time = time
        self.type = _type

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """Variable can never be removed"""
        return False

    def __repr__(self):
        return ("Variable(id={}, activation_id={}, name={}, line={}, type={},"
                "value={})").format(self.id, self.activation_id, self.name,
                                    self.line, self.type, self.value)


class VariableDependencyLW(BaseLW):
    """Variable Dependency lightweight object
    There are type definitions on lightweight.pxd
    """
    __slots__, attributes = define_attrs(
        ["id", "source_activation_id", "source_id",
         "target_activation_id", "target_id", "trial_id", "type"]
    )
    special = set()

    def __init__(self, vid, source_activation_id, source_id,                     # pylint: disable=too-many-arguments
                 target_activation_id, target_id, _type):
        self.id = vid                                                            # pylint: disable=invalid-name
        self.source_activation_id = source_activation_id
        self.source_id = source_id
        self.target_activation_id = target_activation_id
        self.target_id = target_id
        self.trial_id = -1
        self.type = _type

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """Variable Dependency can always be removed"""
        return True

    def __repr__(self):
        return (
            "Dependent(id={}, "
            "sact_id={}, source_id={}, "
            "tact_id={}, target_id={}, type={})"
        ).format(
            self.id,
            self.source_activation_id, self.source_id,
            self.target_activation_id, self.target_id, self.type
        )


class VariableUsageLW(BaseLW):
    """Variable Usage lightweight object
    There are type definitions on lightweight.pxd
    """
    __slots__, attributes = define_attrs(
        ["id", "activation_id", "variable_id",
         "line", "ctx", "trial_id"]
    )
    special = set()

    def __init__(self, vid, activation_id, variable_id, line, ctx):              # pylint: disable=too-many-arguments
        self.id = vid                                                            # pylint: disable=invalid-name
        self.activation_id = activation_id
        self.variable_id = variable_id
        self.line = line
        self.ctx = ctx
        self.trial_id = -1

    def is_complete(self):                                                       # pylint: disable=no-self-use
        """Variable Usage can always be removed"""
        return True

    def __repr__(self):
        return (
            "Usage(id={}, variable_id={}, line={}, ctx={})"
        ).format(self.id, self.variable_id, self.line, self.ctx)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Deployment provenance collector"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import importlib
import modulefinder
import os
import platform
import socket
import sys
import weakref
import getpass
import pkg_resources

from future.utils import viewitems
from future.builtins import map as cvmap

from ...persistence.models import EnvironmentAttr, Module, Dependency
from ...persistence import content
from ...utils.io import print_msg, redirect_output
from ...utils.metaprofiler import meta_profiler
from ...utils.cross_version import string, default_string
from ...utils.functions import version


class Deployment(object):
    """Collect deployment provenance"""

    def __init__(self, metascript):
        self.metascript = weakref.proxy(metascript)

    @meta_profiler("environment")
    def _collect_environment_provenance(self):
        """Collect enviroment variables and operating system characteristics
        Return dict
        """
        attrs = self.metascript.environment_attrs_store

        attrs.add("OS_NAME", platform.system())
        # Unix environment
        try:
            for name in os.sysconf_names:
                try:
                    attrs.add(name, os.sysconf(name))
                except (ValueError, OSError):
                    pass
            for name in os.confstr_names:
                attrs.add(name, os.confstr(name))
        except AttributeError:
            pass

        os_name, _, os_release, os_version, _, _ = platform.uname()
        attrs.add("OS_NAME", os_name)
        attrs.add("OS_RELEASE", os_release)
        attrs.add("OS_VERSION", os_version)

        # Both Unix and Windows
        for attr, value in viewitems(os.environ):
            attrs.add(attr, value)

        attrs.add("USER", getpass.getuser())
        attrs.add("PWD", os.getcwd())
        attrs.add("PID", os.getpid())
        attrs.add("HOSTNAME", socket.gethostname())
        attrs.add("ARCH", platform.architecture()[0])
        attrs.add("PROCESSOR", platform.processor())
        attrs.add("PYTHON_IMPLEMENTATION", platform.python_implementation())
        attrs.add("PYTHON_VERSION", platform.python_version())

        attrs.add("NOWORKFLOW_VERSION", version())

    @meta_profiler("modules")
    def _collect_modules_provenance(self):
        with redirect_output():
            modules = self._find_modules()

            print_msg("  registering provenance from {} modules".format(
                len(modules) - 1))
            self._extract_modules_provenance(modules)

    @meta_profiler("find_modules")
    def _find_modules(self):
        """Use modulefinder to find modules

        Return finder.modules dict
        """
        metascript = self.metascript
        excludes = set()
        last_name = "A" * 255  # invalid name
        max_atemps = 1000
        for i in range(max_atemps):
            try:
                finder = modulefinder.ModuleFinder(excludes=excludes)
                finder.run_script(metascript.path)
                print(metascript.path)
                return finder.modules
            except SyntaxError as exc:
                name = exc.filename.split("site-packages/")[-1]                  # pylint: disable=no-member
                name = name.replace(os.sep, ".")
                name = name[:name.rfind(".")]
                if last_name in name:
                    last_name = last_name[last_name.find(".") + 1:]
                else:
                    last_name = name
                excludes.add(last_name)
                print_msg("   skip module due syntax error: {} ({}/{})"
                          .format(last_name, i + 1, max_atemps))
        return {}

    @meta_profiler("extract_modules")
    def _extract_modules_provenance(self, python_modules):
        """Return a set of module dependencies in the form:
            (name, version, path, code_hash)
        Store module provenance in the content database
        """
        metascript = self.metascript
        modules = metascript.modules_store
        dependencies = metascript.dependencies_store
        modules.id = Module.id_seq()
        for name, module in viewitems(python_modules):
            if name != "__main__":
                module_version = self.get_version(name)
                path = module.__file__
                if path is None:
                    code_hash = None
                else:
                    with open(path, "rb") as fil:
                        code_hash = content.put(fil.read())
                info = (name, module_version, path, code_hash)
                mid = Module.fast_load_module_id(*info) or modules.add(*info)
                dependencies.add(mid)

    def get_version(self, module_name):                                         # pylint: disable=no-self-use
        """Get module version"""
        # Check built-in module
        if module_name in sys.builtin_module_names:
            return platform.python_version()

        # Check package declared module version
        try:
            # ToDo: This is slow! Is there any alternative?
            return pkg_resources.get_distribution(module_name).version
        except Exception:                                                        # pylint: disable=broad-except
            pass

        # Check explicitly declared module version
        try:
            module = importlib.import_module(module_name)
            for attr in ["__version__", "version", "__VERSION__", "VERSION"]:
                try:
                    module_version = getattr(module, attr)
                    if isinstance(module_version, string):
                        return default_string(module_version)
                    if isinstance(module_version, tuple):
                        return ".".join(cvmap(str, module_version))

                except AttributeError:
                    pass
        except Exception:                                                        # pylint: disable=broad-except
            pass

        # If no other option work, return None
        return None

    @meta_profiler("deployment")
    def collect_provenance(self):
        """Collect deployment provenance:
        - environment variables
        - modules dependencies

        metascript should have "trial_id" and "path"
        """
        print_msg("  registering environment attributes")
        self._collect_environment_provenance()

        if self.metascript.bypass_modules:
            print_msg("  using previously detected module dependencies "
                      "(--bypass-modules).")
        else:
            print_msg("  searching for module dependencies")
            self._collect_modules_provenance()

    def store_provenance(self):
        """Store deployment provenance"""
        metascript = self.metascript
        tid = metascript.trial_id
        # Remove after save
        partial = True
        EnvironmentAttr.fast_store(tid, metascript.environment_attrs_store,
                                   partial)
        Module.fast_store(tid, metascript.modules_store, partial)
        Dependency.fast_store(tid, metascript.dependencies_store, partial)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
import sys
import pyposast
from ...now.utils.cross_version import bytes_string
from ...now.collection.prov_definition.slicing_visitor import SlicingVisitor
from ...now.collection.metadata import Metascript


NAME = "noworkflow/tests/examples/script.py"


class TestSlicingDependencies(unittest.TestCase):

        def dependencies(self, line):
            return self.visitor.dependencies[line]

        def call(self, line, col):
            return self.visitor.call_by_col[line][col]

        def parse(self, code):
            metascript = Metascript()
            metascript.fake_path(NAME, bytes_string(code, "utf-8"))
            self.visitor = SlicingVisitor(metascript, metascript.paths[NAME])
            return pyposast.parse(code)

        def test_simple_assignment(self):
            tree = self.parse("a = 1\n"
                              "b = a")
            self.visitor.visit(tree)
            self.assertIn("a", self.dependencies(1))
            self.assertEqual(["a"], self.dependencies(2)["b"])

        def test_unpack_assignment(self):
            tree = self.parse("a, b = (a + c), d")
            self.visitor.visit(tree)
            self.assertEqual(["a", "c"], self.dependencies(1)["a"])
            self.assertEqual(["d"], self.dependencies(1)["b"])

        def test_multiple_assignment(self):
            tree = self.parse("c = a, b = d, e")
            self.visitor.visit(tree)
            self.assertEqual(["d"], self.dependencies(1)["a"])
            self.assertEqual(["e"], self.dependencies(1)["b"])
            self.assertEqual(["d", "e"], self.dependencies(1)["c"])

        def test_for(self):
            tree = self.parse("for i in a + [b, c]:\n"
                              "    pass")
            self.visitor.visit(tree)
            deps = ["a", "b", "c"]
            self.assertEqual([(1, 13)], self.dependencies(1)["i"])
            self.assertEqual(deps, self.dependencies(1)[(1, 13)])


        def test_for2(self):
            tree = self.parse("for i, j in [b, c]:\n"
                              "    pass")
            self.visitor.visit(tree)
            deps = ["b", "c"]
            self.assertEqual([(1, 19)], self.dependencies(1)["i"])
            self.assertEqual([(1, 19)], self.dependencies(1)["j"])
            self.assertEqual(deps, self.dependencies(1)[(1, 19)])

        def test_augmented_assignment(self):
            tree = self.parse("a += 1\n"
                              "b += a")
            self.visitor.visit(tree)
            self.assertEqual(["a"], self.dependencies(1)["a"])
            self.assertEqual(["a", "b"], self.dependencies(2)["b"])

        def test_attribute_assignment(self):
            tree = self.parse("a.attr.attr2 = b")
            self.visitor.visit(tree)
            self.assertEqual(["b"], self.dependencies(1)["a"])
            self.assertEqual(["b"], self.dependencies(1)["a.attr"])
            self.assertEqual(["b"], self.dependencies(1)["a.attr.attr2"])

        def test_subscript_assignment(self):
            tree = self.parse("a[c] = b")
            self.visitor.visit(tree)
            self.assertEqual(["b"], self.dependencies(1)["a"])
            self.assertFalse(self.dependencies(1)["c"])

        def test_complex_subscript_assignment(self):
            tree = self.parse("a.attr.attr5[c.attr2[d.attr3]].attr4 = b")
            self.visitor.visit(tree)
            self.assertEqual(["b"], self.dependencies(1)["a"])
            self.assertEqual(["b"], self.dependencies(1)["a.attr"])
            self.assertEqual(["b"], self.dependencies(1)["a.attr.attr5"])
            self.assertFalse(self.dependencies(1)["a.attr.attr5.attr4"])
            self.assertFalse(self.dependencies(1)["c"])
            self.assertFalse(self.dependencies(1)["c.attr2"])
            self.assertFalse(self.dependencies(1)["d"])
            self.assertFalse(self.dependencies(1)["d.attr3"])

        def test_lambda_assignment(self):
            tree = self.parse("a = (lambda x: x + b)(c)")
            self.visitor.visit(tree)
            self.assertEqual([(1, 24)], self.dependencies(1)["a"])
            self.assertEqual([["c"]], self.call(1, 24).args)

        def test_lambda2_assignment(self):
            tree = self.parse("a = (lambda x: (lambda y: x + y))(b)(c)")
            self.visitor.visit(tree)
            self.assertEqual([(1, 39)], self.dependencies(1)["a"])
            self.assertEqual([["c"]], self.call(1, 39).args)
            self.assertEqual([(1, 36)], self.call(1, 39).func)
            self.assertEqual([["b"]], self.call(1, 36).args)

        def test_list_comprehension_assignment(self):
            tree = self.parse("a = [i + b for i in c if i + d == b]")
            self.visitor.visit(tree)
            if sys.version_info < (3, 0):
                self.assertEqual(["c", "d", "b", "b"],
                                 self.dependencies(1)["a"])
            else:
                self.assertEqual(["c", "d", "b", "b"],
                                 self.dependencies(1)[(1, 36)])
                self.assertEqual([(1, 36)], self.dependencies(1)["a"])

        def test_list_comprehension2_assignment(self):
            tree = self.parse("a = [i + j for i in c for j in d]")
            self.visitor.visit(tree)
            if sys.version_info < (3, 0):
                self.assertEqual(["c", "d"], self.dependencies(1)["a"])
            else:
                self.assertEqual([(1, 33)], self.dependencies(1)["a"])
                self.assertEqual(["c", "d"], self.dependencies(1)[(1, 33)])

        def test_set_comprehension_assignment(self):
            tree = self.parse("a = {i for i in c}")
            self.visitor.visit(tree)
            self.assertEqual([(1, 18)], self.dependencies(1)["a"])
            self.assertEqual(["c"], self.dependencies(1)[(1, 18)])

        def test_generator_assignment(self):
            tree = self.parse("a = sum(i for i in c)")
            self.visitor.visit(tree)
            self.assertEqual((1, 21), self.dependencies(1)["a"][0])
            self.assertEqual(["sum"], self.dependencies(1)[(1, 21)])

        def test_dict_comprehension_assignment(self):
            tree = self.parse("a = {i:i**b for i in c}")
            self.visitor.visit(tree)
            self.assertEqual([(1, 23)], self.dependencies(1)["a"])
            self.assertEqual(["c", "b"], self.dependencies(1)[(1, 23)])

        def test_function_call_assignment(self):
            tree = self.parse("a = fn(b=c)")
            self.visitor.visit(tree)
            self.assertEqual((1, 11), self.dependencies(1)["a"][0])
            self.assertEqual({"b": ["c"]}, self.call(1, 11).keywords)

        def test_nested_call(self):
            tree = self.parse("a = fn(fn(x))")
            self.visitor.visit(tree)
            self.assertEqual((1, 13), self.dependencies(1)["a"][0])
            self.assertEqual([[(1, 12)]], self.call(1, 13).args)

        def test_while(self):
            tree = self.parse("while i:\n"
                              "    b = c\n"
                              "else:\n"
                              "    c = d")
            self.visitor.visit(tree)
            self.assertEqual(["c", "i"], self.dependencies(2)["b"])
            self.assertEqual(["d", "i"], self.dependencies(4)["c"])

        def test_if(self):
            tree = self.parse("if i:\n"
                              "    b = c\n"
                              "else:\n"
                              "    c = d")
            self.visitor.visit(tree)
            self.assertEqual(["c", "i"], self.dependencies(2)["b"])
            self.assertEqual(["d", "i"], self.dependencies(4)["c"])

        def test_nested(self):
            tree = self.parse("if i:\n"
                              "    if j:\n"
                              "        a = x\n"
                              "    b = y\n"
                              "c = z")
            self.visitor.visit(tree)
            self.assertEqual({"x", "i", "j"}, set(self.dependencies(3)["a"]))
            self.assertEqual(["y", "i"], self.dependencies(4)["b"])
            self.assertEqual(["z"], self.dependencies(5)["c"])

        def test_for_independent(self):
            tree = self.parse("for i in a:\n"
                              "    b = c")
            self.visitor.visit(tree)
            self.assertEqual(["c"], self.dependencies(2)["b"])

        def test_for_dependent(self):
            tree = self.parse("for i in a:\n"
                              "    b = b + c")
            self.visitor.visit(tree)
            self.assertEqual(["b", "c", "i"], self.dependencies(2)["b"])

        def test_for_dependent_augment(self):
            tree = self.parse("for i in a:\n"
                              "    b += c")
            self.visitor.visit(tree)
            self.assertEqual(["c", "b", "i"], self.dependencies(2)["b"])

        def test_return(self):
            code = ("def f(a):\n"
                    "    return a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertIn("f", self.dependencies(1))
            self.assertEqual(["a"], self.dependencies(2)["return"])

        def test_yield(self):
            code = ("def f(a):\n"
                    "    yield a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertIn("f", self.dependencies(1))
            self.assertEqual(["a"], self.dependencies(2)["yield"])

        def test_return2(self):
            code = ("def f(a):\n"
                    "    if a:\n"
                    "        return 2")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertIn("f", self.dependencies(1))
            self.assertEqual(["a"], self.dependencies(3)["return"])

        def test_return3(self):
            code = ("def f(a):\n"
                    "    if a:\n"
                    "        return")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertIn("f", self.dependencies(1))
            self.assertEqual(["a"], self.dependencies(3)["return"])

        def test_dec1(self):
            code = ("@decorator\n"
                    "def f(a):\n"
                    "    return a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(["decorator"], self.dependencies(1)[(1, 1)])
            self.assertEqual([(1, 1)], self.dependencies(1)["f"])
            self.assertEqual(["a"], self.dependencies(3)["return"])

        def test_dec2(self):
            code = ("@decorator2\n"
                    "@decorator\n"
                    "def f(a):\n"
                    "    return a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(["decorator2"], self.dependencies(1)[(1, 1)])
            self.assertEqual(["decorator"], self.dependencies(2)[(2, 1)])
            self.assertEqual([(2, 1)], self.dependencies(1)["f"])
            self.assertEqual(["a"], self.dependencies(4)["return"])

        def test_dec3(self):
            code = ("@decorator(x)\n"
                    "def f(a):\n"
                    "    return a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual([["x"]], self.call(1, 13).args)
            self.assertEqual([(1, 13)], self.dependencies(1)[(1, 1)])
            self.assertEqual([(1, 13)], self.call(1, 1).func)
            self.assertEqual([(1, 1)], self.dependencies(1)["f"])
            self.assertEqual(["a"], self.dependencies(3)["return"])

        def test_class1(self):
            code = ("class Test:\n"
                    "    pass")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertIn((1, 5), self.dependencies(1))

        def test_class2(self):
            code = ("@decorator\n"
                    "class Test:\n"
                    "    pass")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(["decorator"], self.dependencies(1)[(1, 1)])
            self.assertEqual([(1, 1)], self.dependencies(1)[(2, 5)])

        def test_class3(self):
            code = ("class Test(Super):\n"
                    "    pass")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(["Super"], self.dependencies(1)[(1, 5)])

        def test_import1(self):
            code = ("import csv\n")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual([(1, 6)], self.dependencies(1)["csv"])

        def test_import2(self):
            code = ("from sys import argv\n")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual([(1, 4)], self.dependencies(1)["argv"])




import csv
import sys
import matplotlib.pyplot as plt
from simulator import simulate

def run_simulation(data_a, data_b):
    return simulate(csv_read(data_a), csv_read(data_b))

def csv_read(f):
    return list(csv.reader(open(f, 'rU'), delimiter=':'))

def extract_column(data, column):
    return [float(row[column]) for row in data]

def plot(data):
    #GetTemperature
    #GetPrecipitation
    plt.scatter(extract_column(data, 0), extract_column(data, 1), marker='o')
    plt.xlabel('Temperature')
    plt.ylabel('Precipitation')
    plt.savefig("output.png")

#Main Program
plot(run_simulation(sys.argv[1], sys.argv[2]))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now diff' module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from future.utils import viewitems, viewkeys

from ..ipython.converter import create_ipynb
from ..persistence.models.diff import Diff as DiffModel
from ..persistence import persistence_config
from ..utils.io import print_msg
from ..utils.cross_version import zip_longest

from .cmd_show import print_trial_relationship
from .command import NotebookCommand


def print_diff_trials(diff):
    """Print diff of basic trial information"""
    for key, values in viewitems(diff.trial):
        print("  {} changed from {} to {}".format(
            key.capitalize().replace("_", " "),
            values[0] or "<None>", values[1] or "<None>"))
    print()


def print_replaced_attributes(replaced, ignore=("id",), extra=tuple(),
                              names=None):
    """Print attributes diff"""
    names = names or {}
    for (removed, added) in replaced:
        print("  Name: {}".format(removed.name))
        output = []
        for key in viewkeys(removed.to_dict(ignore=ignore, extra=extra)):
            removed_attr = getattr(removed, key)
            added_attr = getattr(added, key)
            if removed_attr != added_attr:
                output.append("    {} changed from {} to {}".format(
                    names.get(key, key.capitalize().replace("_", " ")),
                    removed_attr or "<None>", added_attr or "<None>"))
        print("\n".join(output))
        print()


def print_replaced_environment(replaced):
    """Print environment diff"""
    for (attr_removed, attr_added) in replaced:
        print("  Environment attribute {} changed from {} to {}".format(
            attr_removed.name,
            attr_removed.value or "<None>",
            attr_added.value or "<None>"))


def print_brief(added, removed, replaced):
    """Print brief diff"""
    added_names = [access.brief for access in added]
    removed_names = [access.brief for access in removed]
    replaced_names = [rem.brief for rem, add in replaced]
    order = added_names, removed_names, replaced_names
    for column in order:
        column.sort()
    added_names.insert(0, "[Additions]")
    removed_names.insert(0, "[Removals]")
    replaced_names.insert(0, "[Changes]")
    max_column_len = [max(len(text) for text in column)
                      for column in order]

    #max_len = max(len(column) for column in order)
    for add, rem, cha in zip_longest(*order):
        add = add or ""
        rem = rem or ""
        cha = cha or ""
        print('{0: <{3}} | {1: <{4}} | {2: <{5}}'.format(
            add, rem, cha, *max_column_len))


class Diff(NotebookCommand):
    """Compare the collected provenance of two trials"""

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("trial1", type=str,
                help="first trial id to be compared")
        add_arg("trial2", type=str,
                help="second trial id to be compared")
        add_arg("-m", "--modules", action="store_true",
                help="compare module dependencies")
        add_arg("-e", "--environment", action="store_true",
                help="compare environment conditions")
        add_arg("-f", "--file-accesses", action="store_true",
                help="compare read/write access to files")
        add_arg("--brief", action="store_true",
                help="display a concise version of diff")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())

        diff = DiffModel(args.trial1, args.trial2)

        print_msg("trial diff:", True)
        print_diff_trials(diff)

        if args.modules:
            (added, removed, replaced) = diff.modules
            if args.brief:
                print_msg("Brief modules diff", True)
                print_brief(added, removed, replaced)
            else:
                print_msg("{} modules added:".format(len(added)), True)
                print_trial_relationship(added)
                print()

                print_msg("{} modules removed:".format(len(removed)), True)
                print_trial_relationship(removed)
                print()

                print_msg("{} modules replaced:".format(len(replaced)), True)
                print_replaced_attributes(replaced)
            print()

        if args.environment:
            (added, removed, replaced) = diff.environment
            if args.brief:
                print_msg("Brief environment attributes diff", True)
                print_brief(added, removed, replaced)
            else:
                print_msg("{} environment attributes added:".format(
                    len(added)), True)
                print_trial_relationship(added, breakline="\n", other="\n  ")
                print()

                print_msg("{} environment attributes removed:".format(
                    len(removed)), True)
                print_trial_relationship(removed, breakline="\n", other="\n  ")
                print()

                print_msg("{} environment attributes replaced:".format(
                    len(replaced)), True)
                print_replaced_environment(replaced)
            print()

        if args.file_accesses:
            (added, removed, replaced) = diff.file_accesses
            if args.brief:
                print_msg("Brief file access diff", True)
                print_brief(added, removed, replaced)
            else:
                print_msg("{} file accesses added:".format(
                    len(added)), True)
                print_trial_relationship(added)
                print()

                print_msg("{} file accesses removed:".format(
                    len(removed)), True)
                print_trial_relationship(removed)
                print()

                print_msg("{} file accesses replaced:".format(
                    len(replaced)), True)
                print_replaced_attributes(
                    replaced,
                    extra=("mode", "buffering", "content_hash_before",
                           "content_hash_after", "timestamp", "stack"),
                    ignore=("id", "trial_id", "function_activation_id"),
                    names={"stack": "Function"})



    def execute_export(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        DiffModel(args.trial1, args.trial2)
        name = "Diff-{0}-{1}.ipynb".format(args.trial1, args.trial2)
        code = ("%load_ext noworkflow\n"
                "import noworkflow.now.ipython as nip\n"
                "# <codecell>\n"
                "diff = nip.Diff('{0}', '{1}')\n"
                "# diff.graph.view = 0\n"
                "# diff.graph.mode = 3\n"
                "# diff.graph.width = 500\n"
                "# diff.graph.height = 500\n"
                "# <codecell>\n"
                "diff").format(args.trial1, args.trial2)
        create_ipynb(name, code)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Relational Database"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import threading

from os.path import join, exists

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

from ..utils.io import print_msg


DB_FILENAME = "db.sqlite"


class RelationalDatabase(object):
    """Relational Database deal with SQLite connection"""

    def __init__(self, persistence_config):
        self.db_path = None  # Database path
        self.engine = None
        self._session_map = {}
        self.session_factory = sessionmaker()

        self.base = declarative_base()

        persistence_config.add(self)

    def set_path(self, config):
        """Set content_path"""
        self.db_path = join(config.provenance_path, DB_FILENAME)

    def mock(self, config):                                                      # pylint: disable=unused-argument
        """Mock path for tests"""
        self.db_path = ""

    def connect(self, config):
        """Create database connection
        If database does not exist, create it as well
        """
        new_db = not exists(self.db_path)

        if config.should_mock:
            new_db, self.db_path = True, ""

        self.engine = create_engine(
            "sqlite://" + ("/" if self.db_path else "") + self.db_path,
            echo=False)
        self.session_factory.configure(bind=self.engine, autoflush=False,
                                       expire_on_commit=True)
        self._session_map = {}

        if new_db:
            print_msg("creating provenance database")
            self.base.metadata.create_all(self.engine)

    def make_session(self):
        """Create thread safe session"""
        return scoped_session(self.session_factory)

    @property
    def session(self):
        """Access session of current thread"""
        ident = threading.current_thread().ident
        if ident not in self._session_map:
            self._session_map[ident] = self.make_session()
            self._session_map[ident].configure(expire_on_commit=False)
        return self._session_map[ident]

    def query(self, text):
        """Perform SQL query"""
        return self.session.execute(text).fetchall()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ..now.persistence import persistence_config

persistence_config.mock()
persistence_config.connect(".")


from .prov_definition import TestSlicingDependencies
from .prov_execution import TestCallSlicing
from .prov_deployment import TestProvDeployment
from .cross_version_test import TestCrossVersion
from .formatter_test import TestFormatter



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Dependency Filter"""

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref

from collections import defaultdict

from future.utils import viewitems, viewkeys, viewvalues

from .. import Activation, Variable, VariableDependency, FileAccess
from .. import UniqueFileAccess


class ActivationCluster(object):                                                 # pylint: disable=too-few-public-methods
    """Represent an activation cluster"""
    def __init__(self, aid, name, depth=1):
        self.components = []
        self.name = name
        self.same_rank = []
        self.depth = depth
        self.activation_id = aid


def variable_id(variable):
    """Return variable identification for .dot file"""
    if isinstance(variable, FileAccess):
        return "a_{}".format(variable.id)
    act_id = variable.activation_id
    act_id = "global" if act_id == -1 else act_id
    return "v_{}_{}".format(act_id, variable.id)


def escape(string, size=55):
    """Escape string for dot file"""
    if not size or not string:
        return ""
    if len(string) > size:
        half_size = (size - 5) // 2
        string = string[:half_size] + " ... " + string[-half_size:]
    return "" if string is None else string.replace('"', '\\"')


def fix_value(synonym, final):
    """Fix value for synonym"""
    if not synonym.value or synonym.value == "now(n/a)":
        return
    if final.value and final.value != "now(n/a)":
        return
    final.value = synonym.value


class DependencyConfig(object):                                                  # pylint: disable=too-many-instance-attributes
    """Configure dependency graph"""

    def __init__(self):
        self.show_blackbox_dependencies = False
        self.rank_line = True
        self.show_accesses = True
        self.combine_accesses = True
        self.show_external_files = False
        self.max_depth = float("inf")
        self.show_internal_use = True
        self.mode = "simulation"

    @classmethod
    def create_arguments(cls, add_arg, mode="prospective"):
        """Create arguments

        Arguments:
        add_arg -- add argument function
        """
        add_arg("-a", "--accesses", type=int, default=1, metavar="A",
                help="R|show file accesses (default: 1)\n"
                     "0 hides file accesses\n"
                     "1 shows each file once (hide external accesses)\n"
                     "2 shows each file once (show external accesses)\n"
                     "3 shows all accesses (except external accesses)\n"
                     "4 shows all accesses (including external accesses)")
        add_arg("-d", "--depth", type=int, default=0, metavar="D",
                help="R|visualization depth (default: 0)\n"
                     "0 represents infinity")
        add_arg("-u", "--show-internal-use", action="store_false",
                help="show variables and functions which name starts with a "
                     "leading underscore")
        add_arg("-l", "--rank-line", action="store_true",
                help="R|align variables of a line in the same column\n"
                     "With this option, all variables in a loop appear\n"
                     "grouped, reducing the width of the graph.\n"
                     "It may affect the graph legibility.\n"
                     "The alignment is independent for each activation.\n")
        add_arg("-m", "--mode", type=str, default=mode,
                choices=["simulation", "prospective", "dependency"],
                help=("R|Graph mode (default: {})\n"
                      "'simulation' presents a dataflow graph with all\n"
                      "relevant variables.\n"
                      "'prospective' presents only parameters, calls, and\n"
                      "assignments to calls.\n"
                      "'dependency' presents all dependencies, and ignores\n"
                      "depth, rank-line, and hide-accesses configurations"
                      .format(mode)))
        add_arg("-b", "--black-box", action="store_true",
                help="R|propagate black-box dependencies. \n"
                     "Use this option to avoid false negatives. \n"
                     "It shows all dependencies between calls that we do\n"
                     "do not have definitions and that could change some \n"
                     "program state, creating an implicit dependency")


    def read_args(self, args):
        """Read config from args"""
        self.show_accesses = bool(args.accesses)
        self.combine_accesses = args.accesses in {1, 2}
        self.show_external_files = args.accesses in {2, 4}
        self.max_depth = args.depth or float("inf")
        self.show_internal_use = not bool(args.show_internal_use)
        self.rank_line = bool(args.rank_line)
        self.mode = args.mode
        self.show_blackbox_dependencies = bool(args.black_box)


class DependencyFilter(object):                                                  # pylint: disable=too-many-instance-attributes
    """Handle Dot export"""

    def __init__(self, trial):
        super(DependencyFilter, self).__init__()
        self.trial = weakref.proxy(trial)
        self.config = self.trial.dependency_config

        self.valid_types = {"call", "normal", "virtual", "param"}
        self.created = set()
        self.synonyms = {}
        self.departing_arrows = {}
        self.arriving_arrows = {}
        self.variables = {}
        self.accesses = {}

        self.main_cluster = None
        self.current_cluster = None
        self.filtered_variables = set()
        self.dependencies = {}
        self.executed = False

    def _add_variable(self, variable, cluster=None):
        """Create variable in cluster

        Arguments:
        variable -- Variable
        cluster -- activation cluster
        """
        if cluster is None:
            cluster = self.current_cluster
        if variable.name.startswith("_") and not self.config.show_internal_use:
            return
        self.filtered_variables.add(variable_id(variable))
        cluster.components.append(variable)
        self.created.add(variable)

    def _add_all_variables(self, variables, cluster=None):
        """Create all variables"""
        created = self.created
        synonyms = self.synonyms

        for variable in variables:
            variable = synonyms.get(variable, variable)
            if not variable in created and variable.type in self.valid_types:
                self._add_variable(variable, cluster)

    def _all_accesses(self, activation, depth):
        """Get all file accesses recursively if it reaches the maximum depth"""
        for access in activation.file_accesses:
            if self.config.show_external_files or access.is_internal:
                yield access
        if depth + 1 > self.config.max_depth:
            for act in activation.children:
                for access in self._all_accesses(act, depth + 1):
                    yield access

    def _add_call(self, variable, cluster, recursive_function):
        """Check if call is valid for subcluster
        If it is, create a subcluster and call <recursive_function>

        Return return_ variable and type

        There are five possible type:
        fake -- indicates that variable is a fake call (no cluster)
        c_call -- indicates that variable is a c_call (no cluster)
        just_return -- indicates that call has only a return node (no cluster)
        max_depth -- indicates that it reached the max depth (no cluster)
        subgraph -- user defined call within depth (create cluster)
        """
        accesses = self.accesses
        return_ = variable.return_dependency
        if not return_:
            # Fake call
            return None, "fake"
        activation_id = variable.activation_id
        new_activation_id = return_.activation_id
        if self.config.show_accesses:
            for access in self._all_accesses(return_.activation, cluster.depth):
                access = UniqueFileAccess(access._alchemy_pk)
                if (not self.config.combine_accesses or
                        access.name not in accesses):
                    access.value = ""
                    access.line = ""
                    accesses[access.name] = access
                    fcluster = self.main_cluster
                    if not access.is_internal:
                        fcluster = cluster

                    self._add_variable(access, fcluster)
                else:
                    access = accesses[access.name]
                if set("r+") & set(access.mode):
                    self.departing_arrows[variable][access] = "dashed"
                    self.arriving_arrows[access][variable] = "dashed"
                if set("wxa+") & set(access.mode):
                    self.arriving_arrows[variable][access] = "dashed"
                    self.departing_arrows[access][variable] = "dashed"
        if new_activation_id == activation_id:
            # c_call
            variable.value = return_.value
            self.synonyms[return_] = variable
            fix_value(return_, variable)
            return None, "c_call"
        if len(list(return_.activation.variables)) == 1:
            # Just return. Maybe c_call
            variable.value = return_.value
            self.synonyms[return_] = variable
            fix_value(return_, variable)
            return None, "just_return"
        if cluster.depth + 1 > self.config.max_depth:
            # max depth
            variable.value = return_.value
            self.synonyms[return_] = variable
            fix_value(return_, variable)
            return None, "max_depth"
        trial_id = variable.trial_id
        new_activation = Activation((trial_id, new_activation_id))
        ncluster = self.current_cluster = ActivationCluster(
            new_activation_id, variable.name, cluster.depth + 1
        )
        cluster.components.append(ncluster)

        if any([any(return_.dependencies_as_source),
                any(variable.dependencies_as_target)]):

            self._add_variable(return_, ncluster)
            self.synonyms[variable] = return_
            fix_value(variable, return_)

        self._add_all_variables(new_activation.param_variables, ncluster)

        recursive_function(new_activation, ncluster)
        self._prepare_rank(new_activation, ncluster)
        return return_, "subgraph"

    def _prepare_rank(self, activation, cluster):
        """Group variables by line"""
        if self.config.rank_line:
            created = self.created
            by_line = defaultdict(list)
            for variable in activation.variables:
                if variable in created:
                    by_line[variable.line].append(variable)

            for variables in viewvalues(by_line):
                cluster.same_rank.append(variables)

    def _create_dependencies(self, skip_arg=True):
        """Load dependencies from database into a graph"""
        departing_arrows = self.departing_arrows
        arriving_arrows = self.arriving_arrows
        synonyms = self.synonyms
        variables = self.variables

        for sid, tid in VariableDependency.fast_load_by_trial(self.trial.id):
            osource = variables[sid]
            source = synonyms.get(osource, osource)
            otarget = variables[tid]
            target = synonyms.get(otarget, otarget)
            typ = ""
            if (osource.type == otarget.type == "--blackbox--" and
                    not self.config.show_blackbox_dependencies):
                continue

            if "box--" in target.type:
                typ = "dashed"
            if source != target and (not skip_arg or osource.type != "arg"):
                departing_arrows[source][target] = typ
                arriving_arrows[target][source] = typ

    def _fix_dependencies(self):
        """Propagate dependencies, removing missing nodes"""
        created = self.created
        synonyms = self.synonyms
        arriving_arrows = self.arriving_arrows
        departing_arrows = self.departing_arrows

        removed = (
            set(viewvalues(self.variables))
            - created
            - set(viewkeys(synonyms))
        )
        for variable in removed:
            variable_is_box = "box--" in variable.name
            for source, typ_sv in viewitems(arriving_arrows[variable]):
                if (variable_is_box and "box--" in source.name and
                        not self.config.show_blackbox_dependencies):
                    continue
                for target, typ_vt in viewitems(departing_arrows[variable]):
                    if variable_is_box and source.type == target.type == "arg":
                        continue
                    typ = typ_sv or typ_vt
                    if not typ and not variable_is_box:
                        typ = "dashed"

                    #del arriving_arrows[target][variable]
                    #del departing_arrows[variable][target]
                    departing_arrows[source][target] = typ
                    arriving_arrows[target][source] = typ
            del arriving_arrows[variable]
            for target, typ_vt in viewitems(departing_arrows[variable]):
                if (variable_is_box and "box--" in target.name and
                        not self.config.show_blackbox_dependencies):
                    continue
                for source, typ_sv in viewitems(arriving_arrows[variable]):
                    if variable_is_box and source.type == target.type == "arg":
                        continue
                    typ = typ_sv or typ_vt
                    if not typ and not variable_is_box:
                        typ = "dashed"
                    #del arriving_arrows[variable][source]
                    #del departing_arrows[source][variable]
                    departing_arrows[source][target] = typ
                    arriving_arrows[target][source] = typ
            del departing_arrows[variable]

    def _show_dependencies(self):
        """Show dependencies"""
        created = self.created
        departing_arrows = self.departing_arrows

        self._fix_dependencies()

        for source, targets in viewitems(departing_arrows):
            if source not in created:
                continue
            for target, style in viewitems(targets):
                if target not in created or source == target:
                    continue
                dep = (variable_id(source), variable_id(target))
                self.dependencies[dep] = style

    def _dataflow(self, function):
        """Create dataflow graph"""
        synonyms = self.synonyms
        variables = self.variables
        for activation in self.trial.initial_activations:
            function(activation, self.main_cluster)
            self._prepare_rank(activation, self.main_cluster)

        arg_orginal = Variable.fast_arg_and_original(self.trial.id)
        for arg_id, original_id in arg_orginal:
            synonyms[variables[arg_id]] = variables[original_id]
            fix_value(variables[arg_id], variables[original_id])
            if variables[original_id].type == "arg":
                synonyms[variables[arg_id]] = synonyms[variables[original_id]]
                fix_value(variables[arg_id], synonyms[variables[original_id]])

        self._create_dependencies()
        self._show_dependencies()

    def _simulation_activation(self, activation, cluster):
        """Export simulation activation"""
        for variable in activation.variables:
            if (variable.type == "call" and
                    self._add_call(variable, cluster,
                                   self._simulation_activation)[0]):
                continue
            if variable.type in self.valid_types:
                self._add_variable(variable, cluster)

    def _prospective_activation(self, activation, cluster):
        """Export prospective activation"""
        # ToDo: param dependencies
        for variable in activation.no_param_variables:
            if variable.type == "call":
                return_, mode = self._add_call(variable, cluster,
                                               self._prospective_activation)
                if not return_:
                    self._add_variable(variable, cluster)

                if mode == "fake":
                    box = variable.box_dependency
                    if not box:
                        box = variable
                    self._add_all_variables(box.dependencies, cluster)

                elif mode in ("c_call", "just_return"):
                    return_ = variable.return_dependency
                    box = return_.box_dependency
                    if box:
                        self._add_all_variables(box.dependencies, cluster)

                elif mode == "max_depth":
                    return_ = variable.return_dependency
                    for var in return_.activation.param_variables:
                        self._add_all_variables(var.dependencies, cluster)
                elif mode == "subgraph":
                    for var in return_.activation.param_variables:
                        self._add_all_variables(var.dependencies, cluster)

                self._add_all_variables(variable.dependents, cluster)

    def simulation(self):
        """Create simulation graph"""
        self._dataflow(self._simulation_activation)

    def prospective(self):
        """Create prospective graph"""
        self._dataflow(self._prospective_activation)

    def dependency(self):
        """Create dependency graph"""
        self.valid_types = {
            "call", "normal", "virtual", "param", "import",
            "--blackbox--", "--graybox--"
        }

        for variable in viewvalues(self.variables):
            self._add_variable(variable, self.main_cluster)

        self._create_dependencies(skip_arg=False)
        self._show_dependencies()

    def erase(self):
        """Erase graph"""
        self.valid_types = {"call", "normal", "virtual", "param"}
        self.created = set()
        self.synonyms = {}
        self.departing_arrows = defaultdict(dict)
        self.arriving_arrows = defaultdict(dict)
        self.variables = {v.id: v for v in self.trial.variables}
        self.accesses = {}

        self.main_cluster = ActivationCluster(-1, "main")
        self.current_cluster = self.main_cluster
        self.filtered_variables = set()
        self.dependencies = {}

    def run(self):
        """Filter variables graph according to mode"""
        self.erase()
        getattr(self, self.config.mode)()
        self.executed = True


class ActivationClusterVisitor(object):
    """Base ActivationCluster Visitor"""

    visit_initial = None
    visit_activation = None
    visit_access = None
    visit_variable = None

    def __init__(self, dep_filter):
        self.filter = dep_filter

    def _ranks(self, cluster):
        for variables in cluster.same_rank:
            variables = [variable_id(var) for var in variables
                         if variable_id(var) in self.filter.filtered_variables]
            if variables:
                yield variables

    @property
    def filtered_dependencies(self):
        """Get all dependencies"""
        filtered_variables = self.filter.filtered_variables
        for (source, target), style in viewitems(self.filter.dependencies):
            if source in filtered_variables and target in filtered_variables:
                yield source, target, style

    def visit(self, component):
        """Visit component"""
        if isinstance(component, ActivationCluster):
            if component.activation_id == -1 and self.visit_initial:
                return self.visit_initial(component, self._ranks(component))     # pylint: disable=not-callable
            if self.visit_activation:
                return self.visit_activation(component, self._ranks(component))  # pylint: disable=not-callable
        else:
            if not variable_id(component) in self.filter.filtered_variables:
                return
            if isinstance(component, FileAccess) and self.visit_access:
                return self.visit_access(component)                              # pylint: disable=not-callable
            if self.visit_variable:
                return self.visit_variable(component)                            # pylint: disable=not-callable
        return self.visit_default(component)

    def visit_default(self, component):
        """Visit default"""
        pass


class DotVisitor(ActivationClusterVisitor):
    """Create dot file"""

    def __init__(self, fallback, name_length, value_length, types, dep_filter):  # pylint: disable=too-many-arguments
        super(DotVisitor, self).__init__(dep_filter)
        self.result = []
        self.depth = 0
        self.name_length = name_length
        self.value_length = value_length
        self.fallback = fallback
        self.types = types

    def visit_initial(self, cluster, ranks):
        """Visit initial activation"""
        self.result.append("digraph dependency {")
        self.result.append("    rankdir=RL;")
        self.result.append("    node[fontsize=20]")

        for component in cluster.components:
            self.depth = cluster.depth
            self.visit(component)

        for rank in ranks:
            self.result.append(
                "    " * cluster.depth + "{rank=same " + " ".join(rank) + "}"
            )

        for source, target, style in self.filtered_dependencies:
            self.result.append(('    {} -> {} [style="{}"];').format(
                source, target, style
            ))

        self.result.append("}")

    def visit_activation(self, cluster, ranks):
        """Visit other activations"""
        self.result.append(
            "    " * (cluster.depth - 1) +
            "subgraph cluster_{}  {{".format(cluster.activation_id)
        )
        self.result.append("    " * cluster.depth + 'color="#3A85B9";')
        self.result.append("    " * cluster.depth + 'fontsize=30;')
        self.result.append("    " * cluster.depth +
                           'label = "{}";'.format(cluster.name))

        for component in cluster.components:
            self.depth = cluster.depth
            self.visit(component)

        for rank in ranks:
            self.result.append(
                "    " * cluster.depth + "{rank=same " + " ".join(rank) + "}"
            )

        self.result.append("    " * (cluster.depth - 1) + "}")

    def visit_variable(self, variable):
        """Create variable for graph
        Arguments:
        variable -- Variable or FileAccesss object
        depth -- depth for configuring spaces in subclusters
        config -- color schema
        """
        color, shape, font = self._schema_config(variable)
        var = variable_id(variable)

        value = escape(variable.value, self.value_length)
        name = escape(variable.name, self.name_length)

        if value == "now(n/a)":
            value = ""

        label_list = []
        if variable.line:
            label_list.append("{} ".format(variable.line))
        label_list.append(name)
        if value:
            label_list.append(" =\n{}".format(value))
        label = "".join(label_list)

        self.result.append("    " * self.depth + (
            '{var} '
            '[label="{label}"'
            ' fillcolor="{color}" fontcolor="{font}"'
            ' shape="{shape}"'
            ' style="filled"];'
        ).format(var=var, label=label, color=color, font=font, shape=shape))

    def _schema_config(self, variable):
        """Return color schema for variable
        or fallback if there is no valid schema
        """
        if isinstance(variable, FileAccess):
            return self.types.get("access") or self.fallback
        return self.types.get(variable.type) or self.fallback


class PrologVisitor(ActivationClusterVisitor):
    """Export prolog dependencies"""

    def __init__(self, dep_filter):
        super(PrologVisitor, self).__init__(dep_filter)
        self.variables = []
        self.variable_dependencies = []

    @property
    def usages(self):
        """Variable usages generator"""
        filtered_variables = self.filter.filtered_variables
        for usage in self.filter.trial.variable_usages:
            variable = usage.variable
            if variable_id(variable) in filtered_variables:
                yield usage
            else:
                variable = self.filter.synonyms.get(variable, variable)
                if variable_id(variable) in filtered_variables:
                    yield FakeVariableUsage(usage, variable)

    @property
    def dependencies(self):
        """Variable dependencies generator"""
        did = 1
        for source, target, _ in self.filtered_dependencies:
            yield FakeVariableDependency(
                self.filter.trial.id, did, source, target
            )
            did += 1

    def visit_activation(self, cluster, _):
        """Visit activations"""
        for component in cluster.components:
            self.visit(component)

    def visit_access(self, access):
        """Visit access"""
        self.variables.append(FakeVariableAccess(
            access.trial_id, access.id, access.name, access.timestamp
        ))

    def visit_variable(self, variable):
        """Visit variable"""
        self.variables.append(variable)


class FakeVariableAccess(object):
    """Access that mimics variable for variable prolog description"""

    def __init__(self, trial_id, aid, name, timestamp):
        self.trial_id = trial_id
        self.activation_id = "a"
        self.id = "f{}".format(aid)
        self.name = name
        self.line = "nil"
        self.value = ""
        self.time = timestamp


class FakeVariableDependency(object):
    """Access that mimics dependency prolog description"""

    def __init__(self, trial_id, did, source, target):
        self.trial_id = trial_id
        self.id = did
        splitted_source = source.split("_")
        splitted_target = target.split("_")

        self.source_activation_id = splitted_source[-2]
        self.source_id = splitted_source[-1]
        if self.source_activation_id == "a":
            self.source_id = "f{}".format(self.source_id)

        self.target_activation_id = splitted_target[-2]
        self.target_id = splitted_target[-1]
        if self.target_activation_id == "a":
            self.target_id = "f{}".format(self.target_id)


class FakeVariableUsage(object):
    """Usage that mimics variable usage"""

    def __init__(self, original_usage, new_var):
        self.trial_id = new_var.trial_id
        self.activation_id = new_var.activation_id
        self.variable_id = new_var.id
        self.id = original_usage.id
        self.line = original_usage.line

        self.variable = new_var



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'%%now_restore' magic"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ...cmd import Restore
from ...persistence import content

from .command import IpythonCommandMagic


class NowRestore(IpythonCommandMagic, Restore):
    """Restore trial and return the script content

    Examples
    --------
    ::
        In [1]: script_content = %now_restore 2 --file
           ...: print(script_content)
        [now] File script2.py from trial 2 restored
        l = range(4)
        c = sum(l)
        print(c)
    """

    def __init__(self, *args, **kwargs):
        super(NowRestore, self).__init__(*args, **kwargs)
        self.args_file = None
        self.trial = None

    def add_arguments(self):
        super(NowRestore, self).add_arguments()
        add_arg = self.add_argument
        #add_arg("--file", action="store_true",
        #        help="""Restores the original script file""")

    def restore_script(self, trial):
        if self.args_file:
            super(NowRestore, self).restore_script(trial)
        self.trial = trial

    def execute(self, func, line, cell, magic_cls):
        _, args = self.arguments(func, line)
        self.args_file, self.trial = args.file, None
        Restore.execute(self, args)
        return content.get(self.trial.code_hash).decode("utf-8")



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'%now_ip' magic"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .command import IpythonCommandMagic


class NowIpython(IpythonCommandMagic):
    """Return the noWorkflow IPython Module

       Equivalent to:
       In [1]: import noworkflow.now.ipython as nip
          ...: nip
    """

    def execute(self, func, line, cell, magic_cls):
        from ... import ipython as nip
        return nip



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Bytecode transformers"""

import array

from collections import Counter
from types import CodeType

from .dis import code_dis_sorted_line, findlinestarts

from ..cross_version import PY3


def reconstruct_lnotab(first_line, offsets, lines):
    """Reconstruct lnotab according to Python/Objects/lnotab_notes.txt


    Arguments:
    first_line -- first_line in code object (co_firstlineno)
    offsets -- sorted offsets list
    lines -- sorted lines list
    """
    current_offset, current_line = 0, first_line
    new_lnotab = []
    for offset, line in zip(offsets, lines):
        new_offset = offset - current_offset
        while new_offset > 255:
            new_lnotab.append(255)
            new_lnotab.append(0)
            new_offset -= 255
        new_lnotab.append(new_offset)
        new_line = line - current_line
        while new_line > 255:
            new_lnotab.append(255)
            new_lnotab.append(0)
            new_line -= 255
        new_lnotab.append(new_line)
        current_offset, current_line = offset, line
    return array.array('B', new_lnotab).tostring()


def original_line_offsets(code, codes, codes_offsets, codes_lines):
    """Find original offsets and lines


    Arguments:
    code -- CodeType object
    codes_offsets -- dict to be populated with code objects mapped to code ids
    codes_offsets -- dict to be populated with offset lists mapped to code ids
    codes_lines -- dict to be populated with line lists mapped to code ids
    """
    codes[id(code)] = code # necessary to keep reference to code objects
    offsets = codes_offsets[id(code)] = []
    lines = codes_lines[id(code)] = []
    # ToDo: submit a patch to Python to optimize findlinestarts with iters
    for addr, lineno in findlinestarts(code):
        offsets.append(addr)
        lines.append(lineno)

    for const in code.co_consts:
        if isinstance(const, CodeType):
            original_line_offsets(const, codes, codes_offsets, codes_lines)


def recreate_code(code, codes_offsets, codes_lines):
    """Recreate code with new lnotab


    Arguments:
    code -- original code object
    codes_offsets -- updated list of offsets
    codes_lines -- updated list of lines
    """
    offsets = codes_offsets[id(code)]
    lines = codes_lines[id(code)]
    new_lnotab = reconstruct_lnotab(code.co_firstlineno, offsets, lines)
    new_consts = []
    for const in code.co_consts:
        if isinstance(const, CodeType):
            new_consts.append(recreate_code(const, codes_offsets, codes_lines))
        else:
            new_consts.append(const)

    if PY3:
        new_code = CodeType(
            code.co_argcount, code.co_kwonlyargcount, code.co_nlocals,
            code.co_stacksize, code.co_flags, code.co_code, tuple(new_consts),
            code.co_names, code.co_varnames, code.co_filename, code.co_name,
            code.co_firstlineno, new_lnotab, code.co_freevars, code.co_cellvars
        )
    else:
        new_code = CodeType(
            code.co_argcount, code.co_nlocals,
            code.co_stacksize, code.co_flags, code.co_code, tuple(new_consts),
            code.co_names, code.co_varnames, code.co_filename, code.co_name,
            code.co_firstlineno, new_lnotab, code.co_freevars, code.co_cellvars
        )
    return new_code


def insert_new_lines(compiled, process):                                         # pylint: disable=too-many-locals
    """Insert new line events to bytecode
    Return line multiplier and new code

    Arguments:
    compiled -- original code
    process -- process(inst, code, offsets, lines) function which should add
               offsets and lines according to inst
    """
    codes = {}
    codes_offsets = {}
    codes_lines = {}

    original_line_offsets(compiled, codes, codes_offsets, codes_lines)

    for code, inst in code_dis_sorted_line(compiled, recurse=True):
        lines = codes_lines[id(code)]
        offsets = codes_offsets[id(code)]
        process(inst, code, offsets, lines)

    # Create fake line numbers: lines should be unique
    new_codes_lines = {}
    size = max(Counter(lines).most_common(1)[0][1]
               for lines in codes_lines.values())

    for code_id, lines in codes_lines.items():
        new_lines = []
        last = -1
        for line in lines:
            if line != last:
                offset = 0
            new_lines.append(line * size + offset)
            offset += 1
            last = line
        new_codes_lines[code_id] = new_lines

    return size, recreate_code(compiled, codes_offsets, new_codes_lines)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Cross version dis"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import dis as _dis
import sys
import types
from collections import OrderedDict

from .interpreter import InstructionInterpreter


HAVE_CODE = (types.MethodType, types.FunctionType, types.CodeType, type)
PY3 = sys.version_info >= (3, 0)


def _try_compile(source, name, tries=None, compiler=None):
    """Atempt to compile with <compiler> for <tries> modes"""
    if tries is None:
        tries = ["eval", "exec"]
    if not tries:
        raise ValueError("syntax error in passed string")
    try:
        if not compiler:
            return compile(source, name, tries[0])
        else:
            return compiler(source, name, tries[0])
    except SyntaxError:
        return _try_compile(source, name, tries, compiler)


def _get_code_object(obj, compiler=None):
    """Return code object"""
    if isinstance(obj, types.FrameType):  # Frame
        return ("code", obj.f_code)
    if not PY3 and isinstance(obj, types.InstanceType):  # Instance              # pylint: disable=no-member
        obj = obj.__class__
    if hasattr(obj, "__func__"):  # Method
        obj = obj.__func__
    if hasattr(obj, "gi_code"):  # Generator
        obj = obj.gi_code
    if hasattr(obj, "im_func"):  # Function Python 2
        obj = obj.im_func
    if hasattr(obj, "func_code"):  # Function Python 2
        obj = obj.func_code
    if hasattr(obj, "__code__"):  # Function
        obj = obj.__code__
    if isinstance(obj, str):  # Soruce code
        obj = _try_compile(obj, "<string>", compiler=compiler)
    if hasattr(obj, "co_code"):  # Code
        return ("code", obj)
    if hasattr(obj, "__dict__"):  # Class or module
        return ("dict", obj)
    if PY3 and isinstance(obj, (bytes, bytearray)):  # Raw bytecode
        return ("bytes", obj)
    raise TypeError("get_code_object() can not handle '{}' objects".format(
        type(obj).__name__))


def _byte_instructions(code, lasti=-1, varsn=None, names=None, consts=None,      # pylint: disable=too-many-arguments
                       cells=None, linestarts=None, line_offset=0):
    """Generator for byte code instructions
    Check if it starts a line
    """
    interpreter = InstructionInterpreter(code, varsn, names, consts,
                                         cells, linestarts, line_offset)
    # Omit line number
    show_lineno = linestarts is not None
    lineno_width = 3 if show_lineno else 0

    for inst in interpreter:
        inst.lineno_width = lineno_width
        inst.new_source_line = (show_lineno and
                                inst.starts_line is not None and
                                inst.offset > 0)
        inst.is_current_inst = inst.offset == lasti
        yield inst


def _instructions(code, lasti=-1):
    """Generator for code instructions"""
    cell_names = code.co_cellvars + code.co_freevars
    linestarts = OrderedDict(_dis.findlinestarts(code))
    insts = _byte_instructions(
        code.co_code, lasti, code.co_varnames, code.co_names,
        code.co_consts, cell_names, linestarts)

    for inst in insts:
        yield inst


def idis(obj=None, compiler=None, outfile=None):
    """Disassemble objects"""
    typ, code = _get_code_object(obj, compiler)
    if typ == "code":
        for inst in _instructions(code):
            yield inst
    elif typ == "bytes":
        for inst in _byte_instructions(code,):
            yield inst
    elif typ == "dict":
        items = sorted(code.__dict__.items())
        for name, attr in items:
            if isinstance(attr, HAVE_CODE):
                print("Disassembly of %s:" % name, file=outfile)
                try:
                    for inst in idis(attr, compiler=compiler, outfile=outfile):
                        yield inst
                except TypeError as msg:
                    print("Sorry:", msg, file=outfile)
                print(file=outfile)


def _visit(obj, visitor, compiler=None, recurse=False):
    """Recursively disassemble"""
    for _, inst in _visit_with_code(obj, visitor, compiler, recurse):
        yield inst


def _visit_with_code(obj, visitor, compiler=None, recurse=False):
    """Recursively disassemble"""
    typ, obj = _get_code_object(obj, compiler)
    for inst in visitor(obj, compiler=compiler):
        yield obj, inst
    if typ == "code" and recurse:
        for constant in obj.co_consts:
            if type(constant) is type(obj):
                for code, inst in _visit_with_code(constant, visitor,
                                                   compiler, recurse):
                    yield code, inst


def instruction_dis(compiled, compiler=None, recurse=False):
    """Return dis of compiled code"""
    return list(_visit(compiled, idis, compiler, recurse))


def code_dis(compiled, compiler=None, recurse=False):
    """Return dis of compiled code"""
    return list(_visit_with_code(compiled, idis, compiler, recurse))


def instruction_dis_sorted_by_line(compiled, compiler=None, recurse=False):
    """Return dis sorted by line of compiled code"""
    instructions = instruction_dis(compiled, compiler, recurse)
    return sorted(instructions, key=lambda x: (x.line, x.offset))


def code_dis_sorted_line(compiled, compiler=None, recurse=False):
    """Return dis sorted by line of compiled code"""
    instructions = code_dis(compiled, compiler, recurse)
    return sorted(instructions, key=lambda x: (x[1].line, x[1].offset))


findlinestarts = _dis.findlinestarts                                             # pylint: disable=invalid-name



# Dummy calls for representing openings

def pgmtoppm(atlas_slice):
    result = atlas_slice[:-3] + "ppm"
    with open(atlas_slice, "rb") as aslice, \
            open(result, "w") as gif:
        gif.write(atlas_slice + ".ppm")
    return result

def pnmtojpeg(ppm_slice):
    result = ppm_slice[:-3] + "jpg"
    with open(ppm_slice, "rb") as aslice, \
            open(result, "w") as gif:
        gif.write(ppm_slice + ".jpg")
    return result



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Activation Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from future.builtins import map as cvmap
from sqlalchemy import Column, Integer, Text, TIMESTAMP
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint
from sqlalchemy.orm import backref

from ...utils.prolog import PrologDescription, PrologTrial, PrologTimestamp
from ...utils.prolog import PrologAttribute, PrologRepr, PrologNullable

from .base import AlchemyProxy, proxy_class, one, many_viewonly_ref, many_ref
from .base import backref_one, backref_many, query_many_property
from .object_value import ObjectValue
from .variable_dependency import VariableDependency
from .variable import Variable


@proxy_class
class Activation(AlchemyProxy):
    """Represent a function activation"""
    __tablename__ = "function_activation"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "id"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "caller_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    line = Column(Integer)
    return_value = Column(Text)
    start = Column(TIMESTAMP)
    finish = Column(TIMESTAMP)
    caller_id = Column(Integer, index=True)

    _children = backref("children", order_by="Activation.start")
    caller = one(
        "Activation", remote_side=[trial_id, id],
        backref=_children, viewonly=True
    )

    object_values = many_viewonly_ref("activation", "ObjectValue")
    file_accesses = many_viewonly_ref("activation", "FileAccess")

    variables = many_ref("activation", "Variable")
    variables_usages = many_viewonly_ref("activation", "VariableUsage")
    source_variables = many_viewonly_ref(
        "source_activation", "VariableDependency",
        primaryjoin=((id == VariableDependency.m.source_activation_id) &
                     (trial_id == VariableDependency.m.trial_id)))
    target_variables = many_viewonly_ref(
        "target_activation", "VariableDependency",
        primaryjoin=((id == VariableDependency.m.target_activation_id) &
                     (trial_id == VariableDependency.m.trial_id)))

    trial = backref_one("trial")  # Trial.activations
    children = backref_many("children")  # Activation.caller

    @query_many_property
    def globals(self):
        """Return activation globals as a SQLAlchemy query"""
        return self.object_values.filter(ObjectValue.m.type == "GLOBAL")

    @query_many_property
    def arguments(self):
        """Return activation arguments as a SQLAlchemy query"""
        return self.object_values.filter(ObjectValue.m.type == "ARGUMENT")

    @query_many_property
    def param_variables(self):
        """Return param variables as a SQLAlchemy query"""
        return self.variables.filter(Variable.m.type == "param")

    @query_many_property
    def no_param_variables(self):
        """Return param variables as a SQLAlchemy query"""
        return self.variables.filter(Variable.m.type != "param")

    prolog_description = PrologDescription("activation", (
        PrologTrial("trial_id", link="trial.id"),
        PrologAttribute("id"),
        PrologRepr("name"),
        PrologTimestamp("start"),
        PrologTimestamp("finish"),
        PrologNullable("caller_activation_id", attr_name="caller_id",
                       link="activation.id"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "a function *name* was activated\n"
        "by another function (*caller_activation_id*)\n"
        "and executed during a time period from *start*\n"
        "to *finish*."
    ))

    # ToDo: Improve hash

    def __key(self):
        return (self.trial_id, self.name, self.line)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()                                     # pylint: disable=protected-access

    @property
    def duration(self):
        """Calculate activation duration"""
        return int((self.finish - self.start).total_seconds() * 1000000)

    def show(self, _print=lambda x, offset=0: print(x)):
        """Show object

        Keyword arguments:
        _print -- custom print function (default=print)
        """
        global_vars = list(self.globals)
        if global_vars:
            _print("{name}: {values}".format(
                name="Globals", values=", ".join(cvmap(str, global_vars))))

        arg_vars = list(self.arguments)
        if arg_vars:
            _print("{name}: {values}".format(
                name="Arguments", values=", ".join(cvmap(str, arg_vars))))

        if self.return_value:
            _print("Return value: {ret}".format(ret=self.return_value))

        _show_slicing("Variables:", self.variables, _print)
        _show_slicing("Usages:", self.variables_usages, _print)
        _show_slicing("Dependencies:", self.source_variables, _print)

    def __repr__(self):
        return "Activation({0.trial_id}, {0.id}, {0.name})".format(self)


def _show_slicing(name, query, _print):
    """Show slicing objects"""
    objects = list(query)
    if objects:
        _print(name)
        for obj in objects:
            _print(str(obj), 1)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'%now_set_default' magic"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import re
import argparse

from future.utils import viewkeys

from ...persistence.models import MetaModel

from .command import IpythonCommandMagic


class NowSetDefault(IpythonCommandMagic):
    """Set default values on Models

    Examples
    --------
    ::
        In [1]: %now_set_default graph.width=200 graph.height=200

        In [2]: %now_set_default --model History graph_height=100
    """

    def add_arguments(self):
        super(NowSetDefault, self).add_arguments()
        add_arg = self.add_argument
        add_arg("--model", type=str, default="*",
                choices=["*"] + list(viewkeys(MetaModel.__classes__)),
                help="""specifies the model""")
        add_arg("defaults", nargs=argparse.REMAINDER,
                help="Default assingments. Use the format var=value")

    def execute(self, func, line, cell, magic_cls):
        pattern = re.compile(r"\s*(?P<left>[\w\.]+)\s*=\s*(?P<right>\w+)\s*")
        _, args = self.arguments(func, line)
        for match in pattern.finditer(" ".join(args.defaults)):
            right = match.group("right")
            if right.isdigit():
                right = int(right)
            MetaModel.set_classes_default(match.group("left"), right,
                                          model=args.model)



"""
`%hierarchy` and `%%dot` magics for IPython
===========================================

This extension provides two magics.

First magic is ``%hierarchy``.  This magic command draws hierarchy of
given class or the class of given instance.  For example, the
following shows class hierarchy of currently running IPython shell.::

    %hierarchy get_ipython()


Second magic is ``%%dot``.  You can write graphiz dot language in a
cell using this magic.  Example::

    %%dot -- -Kfdp
    digraph G {
        a->b; b->c; c->d; d->b; d->a;
    }


License for ipython-hierarchymagic
----------------------------------

ipython-hierarchymagic is licensed under the term of the Simplified
BSD License (BSD 2-clause license), as follows:

Copyright (c) 2012 Takafumi Arakaki
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


License for Sphinx
------------------

`run_dot` function and `HierarchyMagic._class_name` method in this
extension heavily based on Sphinx code `sphinx.ext.graphviz.render_dot`
and `InheritanceGraph.class_name`.

Copyright notice for Sphinx can be found below.

Copyright (c) 2007-2011 by the Sphinx team (see AUTHORS file).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

from IPython.core.magic import Magics, magics_class, line_magic, cell_magic
from IPython.core.magic_arguments import (argument, magic_arguments,
                                          parse_argstring)
from IPython.core.display import display_png, display_svg

from sphinx.ext.inheritance_diagram import InheritanceGraph

from future.utils import lmap


def run_dot(code, options=[], format='png'):                                     # pylint: disable=redefined-builtin, dangerous-default-value
    """run_dot"""
    # mostly copied from sphinx.ext.graphviz.render_dot
    import os
    from subprocess import Popen, PIPE
    from sphinx.util.osutil import EPIPE, EINVAL

    dot_args = ['dot'] + options + ['-T', format]
    if os.name == 'nt':
        # Avoid opening shell window.
        # * https://github.com/tkf/ipython-hierarchymagic/issues/1
        # * http://stackoverflow.com/a/2935727/727827
        process = Popen(dot_args, stdout=PIPE, stdin=PIPE, stderr=PIPE,
                        creationflags=0x08000000)
    else:
        process = Popen(dot_args, stdout=PIPE, stdin=PIPE, stderr=PIPE)
    wentwrong = False
    try:
        # Graphviz may close standard input when an error occurs,
        # resulting in a broken pipe on communicate()
        stdout, stderr = process.communicate(code.encode('utf-8'))
    except (OSError, IOError) as err:
        if err.errno != EPIPE:
            raise
        wentwrong = True
    if wentwrong:
        # in this case, read the standard output and standard error streams
        # directly, to get the error message(s)
        stdout, stderr = process.stdout.read(), process.stderr.read()
        process.wait()
    if process.returncode != 0:
        raise RuntimeError('dot exited with error:\n[stderr]\n{0}'
                           .format(stderr.decode('utf-8')))
    return stdout


@magics_class
class GraphvizMagic(Magics):
    """GraphvizMagic"""

    @magic_arguments()
    @argument(
        '-f', '--format', default='png', choices=('png', 'svg'),
        help='output format (png/svg)'
    )
    @argument(
        'options', default=[], nargs='*',
        help='options passed to the `dot` command'
    )
    @cell_magic
    def dot(self, line, cell):
        """Draw a figure using Graphviz dot command."""
        args = parse_argstring(self.dot, line)

        image = run_dot(cell, args.options, format=args.format)

        if args.format == 'png':
            display_png(image, raw=True)
        elif args.format == 'svg':
            display_svg(image, raw=True)


class FoldedInheritanceGraph(InheritanceGraph):
    """InheritanceGraph"""

    def __init__(self, *args, **kwds):
        self._width = kwds.pop('width', 40)
        super(FoldedInheritanceGraph, self).__init__(*args, **kwds)

    @staticmethod
    def _foldclassname(classname, width):
        """Split `classname` in newlines if the width is wider than `width`.

        >>> fold = FoldedInheritanceGraph._foldclassname
        >>> fold('aaa.bbb.ccc', 7)
        'aaa.bbb\\n.ccc'
        >>> fold('aaa.bbb.ccc', 3)
        'aaa\\n.bbb\\n.ccc'
        >>> identity = lambda x, y: ''.join(fold(x, y).split('\\n'))
        >>> identity('aaa.bbb.ccc', 7)
        'aaa.bbb.ccc'
        >>> identity('aaa.bbb.ccc', 3)
        'aaa.bbb.ccc'
        """
        parts = classname.split('.')
        lines = []
        chunk = [parts.pop(0)]
        for part in parts:
            if len('.'.join(chunk + [part])) > width:
                lines.append('.'.join(chunk))
                chunk = [part]
            else:
                chunk.append(part)
        lines.append('.'.join(chunk))
        return '\\n.'.join(lines)

    def _class_info(self, *args, **kwds):
        class_info = super(FoldedInheritanceGraph, self) \
            ._class_info(*args, **kwds)
        width = self._width

        def fold(elem):
            """fold"""
            (nodename, fullname, baselist) = elem
            nodename = self._foldclassname(nodename, width)
            baselist = [self._foldclassname(b, width) for b in baselist]
            return (nodename, fullname, baselist)

        return lmap(fold, class_info)


@magics_class
class HierarchyMagic(Magics):
    """HierarchyMagic"""

    @magic_arguments()
    @argument(
        '-r', '--rankdir', default='TB',
        help='direction of the hierarchy graph (default: %(default)s)'
    )
    @argument(
        '-s', '--size', default='5.0, 12.0',
        help='size of the generated figure (default: %(default)s)',
    )
    @argument(
        '-w', '--name-width', default=40, type=int,
        help='width of each nodes in character length (default: %(default)s)',
    )
    @argument(
        'object', nargs='+',
        help='Class hierarchy of these classes or objects will be drawn',
    )
    @line_magic
    def hierarchy(self, parameter_s=''):
        """Draw hierarchy of a given class."""
        args = parse_argstring(self.hierarchy, parameter_s)
        objects = lmap(self.shell.ev, args.object)
        clslist = lmap(self._object_to_class, objects)
        namelist = lmap(self._class_name, clslist)
        igraph = FoldedInheritanceGraph(
            namelist, '',
            width=args.name_width)
        code = igraph.generate_dot(
            'inheritance_graph',
            graph_attrs={'rankdir': args.rankdir,
                         'size': '"{0}"'.format(args.size)})
        stdout = run_dot(code, format='png')
        display_png(stdout, raw=True)

    @staticmethod
    def _object_to_class(obj):
        if isinstance(obj, type):
            return obj
        elif hasattr(obj, "__class__"):
            return obj.__class__
        else:
            raise ValueError(
                "Given object {0} is not a class or an instance".format(obj))

    @staticmethod
    def _class_name(clas, parts=0):
        """Given a class object, return a fully-qualified name.

        This works for things I've tested in matplotlib so far, but may not be
        completely general.
        """
        module = clas.__module__
        if module == '__builtin__':
            fullname = clas.__name__
        else:
            fullname = '%s.%s' % (module, clas.__name__)
        if parts == 0:
            return fullname
        name_parts = fullname.split('.')
        return '.'.join(name_parts[-parts:])


def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    global _loaded                                                               # pylint: disable=global-statement, invalid-name
    if not _loaded:
        ipython.register_magics(HierarchyMagic)
        ipython.register_magics(GraphvizMagic)
        _loaded = True

_loaded = False                                                                  # pylint: disable=invalid-name



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Head Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text
from sqlalchemy import ForeignKeyConstraint

from ...utils.prolog import PrologDescription, PrologRepr, PrologAttribute

from .. import relational

from .base import AlchemyProxy, proxy_class, one, proxy


@proxy_class
class Head(AlchemyProxy):
    """Represent a Head of a branch"""

    __tablename__ = "head"
    __table_args__ = (
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="SET NULL"),
        {"sqlite_autoincrement": True},
    )
    id = Column(Integer, primary_key=True)                                       # pylint: disable=invalid-name
    script = Column(Text)
    trial_id = Column(Integer, index=True)

    trial = one("Trial")

    prolog_description = PrologDescription("head", (
        PrologRepr("script"),
        PrologAttribute("trial_id"),
    ))

    def __repr__(self):
        return "Head({0.id}, {0.script}, {0.trial_id})".format(self)

    @classmethod  # query
    def load_head(cls, script, session=None):
        """Load head by script"""
        session = session or relational.session
        return proxy(
            session.query(cls.m)
            .filter((cls.m.script == script))
            .first()
        )

    @classmethod  # query
    def remove(cls, hid, session=None):
        """Remove head by id"""
        session = session or relational.session
        head = session.query(cls.m).get(hid)
        session.delete(head)
        session.commit()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Test now.prov_deployment module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
import platform

from future.utils import viewvalues

from ...now.collection.metadata import Metascript


NAME = "./noworkflow/tests/examples/script.py"
MODULES = "./noworkflow/tests/examples/modules.py"


class TestProvDeployment(unittest.TestCase):
    """TestCase for now.prov_deployment module"""

    def prepare(self, name=NAME):
        metascript = Metascript()
        metascript.path = name
        metascript.name = name
        metascript.verbose = True
        return metascript

    def test_collect_environment_provenance(self):
        metascript = self.prepare()
        metascript.deployment._collect_environment_provenance()
        env = {e.name: e.value
               for e in viewvalues(metascript.environment_attrs_store.store)}
        self.assertIn("PWD", env)
        self.assertIn("PYTHON_VERSION", env)
        self.assertIn("PYTHON_IMPLEMENTATION", env)

    def test_collect_modules_provenance(self):
        metascript = self.prepare(name=MODULES)
        metascript.deployment._collect_modules_provenance()
        modules = {m.name
                   for m in viewvalues(metascript.modules_store.store)}
        self.assertIn("ast", modules)
        self.assertIn("script", modules)

    def test_get_version_system_module(self):
        metascript = self.prepare()
        self.assertEqual(platform.python_version(),
                         metascript.deployment._get_version("sys"))

    def test_get_version_distribution(self):
        metascript = self.prepare()
        self.assertEqual("1.1.3",
                         metascript.deployment._get_version("pyposast"))

    def test_get_version_other(self):
        metascript = self.prepare()
        self.assertEqual(None, metascript.deployment._get_version("other"))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define f_trace related interpreters and functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from future.utils import bytes_to_native_str as n

from .code_interpreter import CodeInterpreter, PyInterpreter


class AlmostReadOnlyDict(dict):
    """Use it to avoid changes on the original dict"""

    def __init__(self, *args, **kwargs):
        super(AlmostReadOnlyDict, self).__init__(*args, **kwargs)
        self.other = {}

    def __getitem__(self, item):
        if item in self.other:
            return self.other[item]
        return super(AlmostReadOnlyDict, self).__getitem__(item)

    def __setitem__(self, item, value):
        self.other[item] = value

    def __delitem__(self, item):
        if item in self.other:
            del self.other[item]


class FindFTrace(CodeInterpreter):                                               # pylint: disable=too-many-instance-attributes
    """Find <expr>.f_trace attribution"""

    def __init__(self, *args, **kwargs):
        # Disable operations that may cause effect
        # Default
        # self.store_fast = self.nop
        self.store_subscr = self.nop
        # self.store_name = self.nop
        self.store_global = self.nop
        # self.delete_fast = self.nop
        self.delete_subscr = self.nop
        # self.delete_name = self.nop
        self.delete_attr = self.nop
        self.delete_global = self.nop
        self.print_expr = self.nop

        # Python 2
        self.store_slice__0 = self.nop
        self.store_slice__0 = self.nop
        self.store_slice__1 = self.nop
        self.store_slice__2 = self.nop
        self.store_slice__3 = self.nop
        self.delete_slice__0 = self.nop
        self.delete_slice__1 = self.nop
        self.delete_slice__2 = self.nop
        self.delete_slice__3 = self.nop

        super(FindFTrace, self).__init__(*args, **kwargs)

        self._locals = AlmostReadOnlyDict(self._locals)
        self._globals = AlmostReadOnlyDict(self._globals)

    def store_attr(self):
        """STORE_ATTR opcode"""
        if self.names[self.oparg] == "f_trace":
            self._stop = True
            self.result = self.stack.pop() if self.stack else True


FTraceExe = type(n(b"FTraceExe"), (FindFTrace, PyInterpreter), {})               # pylint: disable=invalid-name


def get_f_trace(code, loc, glob):
    """Get frame from frame.f_trace attribution"""
    interpreter = FTraceExe(code, loc, glob)
    interpreter.execute()
    return interpreter.result


def find_f_trace(code, loc, glob, lasti):
    """Check if code has frame.f_trace attribution"""
    if "f_trace" not in code.co_names:
        return False
    interpreter = FindFTrace(code, loc, glob)
    interpreter.execute()
    if not interpreter.result:
        return False

    last_offset = 0
    for offset in interpreter.linestarts:
        if offset >= interpreter.opi:
            return lasti == last_offset
        last_offset = offset
    return False



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


from noworkflow.now.cmd import main

main()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
import sys

from ...now.cmd.cmd_run import run
from ...now.collection.metadata import Metascript


NAME = "noworkflow/tests/examples/script.py"
PY3_PREFIX = "" if sys.version_info < (3, 0) else "module."


class Args(object):

    def __init__(self):
        self.verbose = False
        self.bypass_modules = False
        self.context = "main"
        self.depth = sys.getrecursionlimit()
        self.non_user_depth = 1
        self.execution_provenance = "Tracer"
        self.disasm = False
        self.dir = None
        self.script = NAME
        self.argv = ["-e", "Tracer", "__init__.py"]
        self.create_last = False
        self.name = None
        self.meta = False
        self.disasm0 = False
        self.disasm = False
        self.save_frequency = 0
        self.call_storage_frequency = 10000


class TestCallSlicing(unittest.TestCase):

    def prepare(self, code):
        sys.argv = ["now", "run", "-e", "Tracer", "__init__.py"]
        metascript = Metascript().read_cmd_args(Args())
        metascript.fake_path(NAME, code.encode("utf-8"))

        # Set __main__ namespace
        import __main__
        metascript.namespace = __main__.__dict__

        # Clear boilerplate
        metascript.clear_sys()
        metascript.clear_namespace()

        return metascript

    def extract(self, metascript):
        result = set()
        for dep in metascript.variables_dependencies_store.values():
            source = metascript.variables_store[dep.source_id]
            target = metascript.variables_store[dep.target_id]
            result.add(((source.name, source.line),
                        (target.name, target.line)))
        return result

    def test_simple(self):
        metascript = self.prepare("def fn(a, b):\n"
                                  "    return a + b\n"
                                  "x = y = 1\n"
                                  "r = fn(x, y)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("b", 1), ("y", 3)),
            (("return", 2), ("a", 1)),
            (("return", 2), ("b", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_call_keyword(self):
        metascript = self.prepare("def fn(a, b, c=3):\n"
                                  "    return a + b + c\n"
                                  "x = y = 1\n"
                                  "r = fn(x, y, c=y)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("b", 1), ("y", 3)),
            (("c", 1), ("y", 3)),
            (("return", 2), ("a", 1)),
            (("return", 2), ("b", 1)),
            (("return", 2), ("c", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_call_keyword_kw(self):
        metascript = self.prepare("def fn(a, b, c=3):\n"
                                  "    return a + b + c\n"
                                  "x = y = 1\n"
                                  "z = {'b': 2}\n"
                                  "r = fn(x, c=y, **z)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("b", 1), ("z", 4)),
            (("c", 1), ("y", 3)),
            (("return", 2), ("a", 1)),
            (("return", 2), ("b", 1)),
            (("return", 2), ("c", 1)),
            (("call fn", 5), ("return", 2)),
            (("call fn", 5), ("fn", 1)),
            (("z", 5), ("x", 3)),
            (("z", 5), ("y", 3)),
            (("z", 5), ("z", 4)),
            (("r", 5), ("call fn", 5)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_call_args(self):
        metascript = self.prepare("def fn(a, b, c=3):\n"
                                  "    return a + b + c\n"
                                  "x = 1\n"
                                  "y = [2]\n"
                                  "r = fn(x, *y)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("b", 1), ("y", 4)),
            (("c", 1), ("y", 4)),  # ToDo #75: fix?
            (("return", 2), ("a", 1)),
            (("return", 2), ("b", 1)),
            (("return", 2), ("c", 1)),
            (("call fn", 5), ("return", 2)),
            (("call fn", 5), ("fn", 1)),
            (("y", 5), ("y", 4)),
            (("y", 5), ("x", 3)),
            (("r", 5), ("call fn", 5)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_def_args(self):
        metascript = self.prepare("def fn(a, *args):\n"
                                  "    return a + args[0]\n"
                                  "x, y, z = 1, 2, 3\n"
                                  "r = fn(x, y, z)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("args", 1), ("y", 3)),
            (("args", 1), ("z", 3)),  # ToDo #75: fix?
            (("return", 2), ("a", 1)),
            (("return", 2), ("args", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_def_args_call_args(self):
        metascript = self.prepare("def fn(a, *args):\n"
                                  "    return a + args[0]\n"
                                  "x, y = 1, [2, 3]\n"
                                  "r = fn(x, *y)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("args", 1), ("y", 3)),
            (("return", 2), ("a", 1)),
            (("return", 2), ("args", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("y", 4), ("y", 3)),
            (("y", 4), ("x", 3)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_def_args_call_kw(self):
        metascript = self.prepare("def fn(a, *args):\n"
                                  "    return a\n"
                                  "x = {'a': 1}\n"
                                  "r = fn(**x)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("args", 1), ("x", 3)),  # ToDo #75: fix?
            (("return", 2), ("a", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("x", 4), ("x", 3)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_def_kwargs_call_keywords(self):
        metascript = self.prepare("def fn(a, **kwargs):\n"
                                  "    return a + kwargs['b']\n"
                                  "x, y = 1, 1\n"
                                  "r = fn(x, b=y)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("kwargs", 1), ("y", 3)),
            (("return", 2), ("a", 1)),
            (("return", 2), ("kwargs", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_def_kwargs_call_kwargs(self):
        metascript = self.prepare("def fn(a, **kwargs):\n"
                                  "    return a + kwargs['b']\n"
                                  "x = {'a': 1, 'b': 2}\n"
                                  "r = fn(**x)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("kwargs", 1), ("x", 3)),
            (("return", 2), ("a", 1)),
            (("return", 2), ("kwargs", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("x", 4), ("x", 3)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_complex(self):
        metascript = self.prepare("def fn(a, b, c, d, "
                                  "e=5, f=6, g=7, **kwargs):\n"
                                  "    return a\n"
                                  "x, y, z, w, u = 1, 2, [3, 4], 5, 7\n"
                                  "v = {'h': 8}\n"
                                  "r = fn(x, y, *z, e=w, g=u, **v)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("b", 1), ("y", 3)),
            (("c", 1), ("z", 3)),
            (("c", 1), ("v", 4)),  # ToDo #75: fix?
            (("d", 1), ("z", 3)),
            (("d", 1), ("v", 4)),  # ToDo #75: fix?
            (("e", 1), ("w", 3)),
            (("f", 1), ("z", 3)),  # ToDo #75: fix?
            (("f", 1), ("v", 4)),  # ToDo #75: fix?
            (("g", 1), ("u", 3)),
            (("kwargs", 1), ("z", 3)),  # ToDo #75: fix?
            (("kwargs", 1), ("v", 4)),
            (("return", 2), ("a", 1)),
            (("call fn", 5), ("return", 2)),
            (("call fn", 5), ("fn", 1)),
            (("z", 5), ("x", 3)),
            (("z", 5), ("y", 3)),
            (("z", 5), ("z", 3)),
            (("z", 5), ("w", 3)),
            (("z", 5), ("u", 3)),
            (("z", 5), ("v", 4)),
            (("v", 5), ("x", 3)),
            (("v", 5), ("y", 3)),
            (("v", 5), ("z", 3)),
            (("v", 5), ("w", 3)),
            (("v", 5), ("u", 3)),
            (("v", 5), ("v", 4)),
            (("r", 5), ("call fn", 5)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_nested(self):
        metascript = self.prepare("def fn(a):\n"
                                  "    return a\n"
                                  "x = 1\n"
                                  "r = fn(fn(x))")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("a", 1), ("call fn", 4)),
            (("return", 2), ("a", 1)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_ccall(self):
        metascript = self.prepare("a, b = 1, 2\n"
                                  "c = min(a, b)")
        run(metascript)
        result = {
            (("return", 2), ("a", 1)),
            (("return", 2), ("b", 1)),
            (("call {}min".format(PY3_PREFIX), 2), ("return", 2)),
            (("call {}min".format(PY3_PREFIX), 2), ("min", 0)),
            (("c", 2), ("call {}min".format(PY3_PREFIX), 2)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_noreturn(self):
        metascript = self.prepare("def fn(a):\n"
                                  "    a = 2\n"
                                  "x = 1\n"
                                  "r = fn(x)")
        run(metascript)
        result = {
            (("a", 1), ("x", 3)),
            (("call fn", 4), ("return", 2)),
            (("call fn", 4), ("fn", 1)),
            (("r", 4), ("call fn", 4)),
        }
        self.assertEqual(result, self.extract(metascript))

    def test_ccall_inter_params(self):
        metascript = self.prepare("a, b = [1, 2, 3], True\n"
                                  "c = sorted(a, reverse=b)")
        run(metascript)
        result = {
            (("return", 2), ("a", 1)),
            (("return", 2), ("b", 1)),
            (("call {}sorted".format(PY3_PREFIX), 2), ("return", 2)),
            (("call {}sorted".format(PY3_PREFIX), 2), ("sorted", 0)),
            (("a", 2), ("a", 1)),
            (("a", 2), ("b", 1)),
            (("c", 2), ("call {}sorted".format(PY3_PREFIX), 2)),
        }
        if sys.version_info < (3, 0):
            result.add(
                (("b", 1), ("True", 0)),
            )
        self.assertEqual(result, self.extract(metascript))

    def test_import1(self):
        metascript = self.prepare("import csv")
        run(metascript)
        result = {
            (("csv", 1), ("call import csv", 1)),
        }
        self.assertEqual(result, self.extract(metascript))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Persistence Models"""
from __future__ import (absolute_import, print_function,
                        division)

from future.utils import text_to_native_str as n

# Base
from .base import Model, MetaModel

# Database Models
from .activation import Activation
from .dependency import Dependency
from .environment_attr import EnvironmentAttr
from .file_access import FileAccess, UniqueFileAccess
from .function_def import FunctionDef
from .graph_cache import GraphCache
from .head import Head
from .module import Module
from .object import Object
from .object_value import ObjectValue
from .variable import Variable
from .variable_dependency import VariableDependency
from .variable_usage import VariableUsage
from .tag import Tag
from .trial import Trial

# Other models
from .history import History
from .diff import Diff
from .trial_prolog import TrialProlog


ORDER = [
    Trial, Head, Tag, GraphCache,  # Trial
    Module, Dependency, EnvironmentAttr,  # Deployment
    FunctionDef, Object,  # Definition
    Activation, ObjectValue, FileAccess,  # Execution
    Variable, VariableUsage, VariableDependency  # Slicing
]


__all__ = [
    n(x.__modelname__) for x in ORDER
] + [
    "History",
    "Diff",
    "TrialProlog",

    "MetaModel",
    "Model",
    "ORDER"
]



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Test now.cross_version module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
from ..now.utils.cross_version import bytes_string, cross_compile


class TestCrossVersion(unittest.TestCase):
    """TestCase for now.cross_version module"""

    def test_bytes_string_unicode_to_bytes(self):
        string = u"\u0435"
        self.assertEqual(1, len(string))
        self.assertEqual(2, len(bytes_string(string)))

    def test_bytes_string_bytes_to_bytes(self):
        string = b"a"
        self.assertEqual(1, len(string))
        self.assertEqual(1, len(bytes_string(string)))

    def test_cross_compile(self):
        code = b"a = 2"
        expected = compile(code, "name", "exec")
        result = cross_compile(code, "name", "exec")
        args = [
            "co_argcount", "co_cellvars", "co_code", "co_consts",
            "co_filename", "co_firstlineno", "co_freevars", "co_lnotab",
            "co_name", "co_names", "co_nlocals", "co_stacksize", "co_varnames"
        ]
        for arg in args:
            self.assertEqual(getattr(expected, arg), getattr(result, arg))
        # On Python 2: result.co_flags != expected.co_flags
